# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from libcloud.utils.py3 import httplib
from libcloud.utils.misc import reverse_dict
from libcloud.common.base import JsonResponse, PollingConnection
from libcloud.common.types import LibcloudError
from libcloud.common.openstack import OpenStackDriverMixin
from libcloud.common.rackspace import AUTH_URL
from libcloud.loadbalancer.base import DEFAULT_ALGORITHM, Driver, Member, Algorithm, LoadBalancer
from libcloud.loadbalancer.types import State, MemberCondition
from libcloud.compute.drivers.rackspace import RackspaceConnection

try:
    import simplejson as json
except ImportError:
    import json


ENDPOINT_ARGS_MAP = {
    "dfw": {
        "service_type": "rax:load-balancer",
        "name": "cloudLoadBalancers",
        "region": "DFW",
    },
    "ord": {
        "service_type": "rax:load-balancer",
        "name": "cloudLoadBalancers",
        "region": "ORD",
    },
    "iad": {
        "service_type": "rax:load-balancer",
        "name": "cloudLoadBalancers",
        "region": "IAD",
    },
    "lon": {
        "service_type": "rax:load-balancer",
        "name": "cloudLoadBalancers",
        "region": "LON",
    },
    "syd": {
        "service_type": "rax:load-balancer",
        "name": "cloudLoadBalancers",
        "region": "SYD",
    },
    "hkg": {
        "service_type": "rax:load-balancer",
        "name": "cloudLoadBalancers",
        "region": "HKG",
    },
}


class RackspaceResponse(JsonResponse):
    def parse_body(self):
        if not self.body:
            return None
        return super().parse_body()

    def success(self):
        return 200 <= int(self.status) <= 299


class RackspaceHealthMonitor:
    """
    :param type: type of load balancer.  currently CONNECT (connection
                 monitoring), HTTP, HTTPS (connection and HTTP
                 monitoring) are supported.
    :type type: ``str``

    :param delay: minimum seconds to wait before executing the health
                      monitor.  (Must be between 1 and 3600)
    :type delay: ``int``

    :param timeout: maximum seconds to wait when establishing a
                    connection before timing out.  (Must be between 1
                    and 3600)
    :type timeout: ``int``

    :param attempts_before_deactivation: Number of monitor failures
                                         before removing a node from
                                         rotation. (Must be between 1
                                         and 10)
    :type attempts_before_deactivation: ``int``
    """

    def __init__(self, type, delay, timeout, attempts_before_deactivation):
        self.type = type
        self.delay = delay
        self.timeout = timeout
        self.attempts_before_deactivation = attempts_before_deactivation

    def __repr__(self):
        return (
            "<RackspaceHealthMonitor: type=%s, delay=%d, timeout=%d, "
            "attempts_before_deactivation=%d>"
            % (self.type, self.delay, self.timeout, self.attempts_before_deactivation)
        )

    def _to_dict(self):
        return {
            "type": self.type,
            "delay": self.delay,
            "timeout": self.timeout,
            "attemptsBeforeDeactivation": self.attempts_before_deactivation,
        }


class RackspaceHTTPHealthMonitor(RackspaceHealthMonitor):
    """
    A HTTP health monitor adds extra features to a Rackspace health monitor.

    :param path: the HTTP path to monitor.
    :type path: ``str``

    :param body_regex: Regular expression used to evaluate the body of
                       the HTTP response.
    :type body_regex: ``str``

    :param status_regex: Regular expression used to evaluate the HTTP
                         status code of the response.
    :type status_regex: ``str``
    """

    def __init__(
        self,
        type,
        delay,
        timeout,
        attempts_before_deactivation,
        path,
        body_regex,
        status_regex,
    ):
        super().__init__(type, delay, timeout, attempts_before_deactivation)
        self.path = path
        self.body_regex = body_regex
        self.status_regex = status_regex

    def __repr__(self):
        return (
            "<RackspaceHTTPHealthMonitor: type=%s, delay=%d, timeout=%d, "
            "attempts_before_deactivation=%d, path=%s, body_regex=%s, "
            "status_regex=%s>"
            % (
                self.type,
                self.delay,
                self.timeout,
                self.attempts_before_deactivation,
                self.path,
                self.body_regex,
                self.status_regex,
            )
        )

    def _to_dict(self):
        super_dict = super()._to_dict()
        super_dict["path"] = self.path
        super_dict["statusRegex"] = self.status_regex

        if self.body_regex:
            super_dict["bodyRegex"] = self.body_regex

        return super_dict


class RackspaceConnectionThrottle:
    """
    :param min_connections: Minimum number of connections per IP address
                            before applying throttling.
    :type min_connections: ``int``

    :param max_connections: Maximum number of connections per IP address.
                            (Must be between 0 and 100000, 0 allows an
                            unlimited number of connections.)
    :type max_connections: ``int``

    :param max_connection_rate: Maximum number of connections allowed
                                from a single IP address within the
                                given rate_interval_seconds.  (Must be
                                between 0 and 100000, 0 allows an
                                unlimited number of connections.)
    :type max_connection_rate: ``int``

    :param rate_interval_seconds: Interval at which the
                                  max_connection_rate is enforced.
                                  (Must be between 1 and 3600.)
    :type rate_interval_seconds: ``int``
    """

    def __init__(
        self,
        min_connections,
        max_connections,
        max_connection_rate,
        rate_interval_seconds,
    ):
        self.min_connections = min_connections
        self.max_connections = max_connections
        self.max_connection_rate = max_connection_rate
        self.rate_interval_seconds = rate_interval_seconds

    def __repr__(self):
        return (
            "<RackspaceConnectionThrottle: min_connections=%d, "
            "max_connections=%d, max_connection_rate=%d, "
            "rate_interval_seconds=%d>"
            % (
                self.min_connections,
                self.max_connections,
                self.max_connection_rate,
                self.rate_interval_seconds,
            )
        )

    def _to_dict(self):
        return {
            "maxConnections": self.max_connections,
            "minConnections": self.min_connections,
            "maxConnectionRate": self.max_connection_rate,
            "rateInterval": self.rate_interval_seconds,
        }


class RackspaceAccessRuleType:
    ALLOW = 0
    DENY = 1

    _RULE_TYPE_STRING_MAP = {ALLOW: "ALLOW", DENY: "DENY"}


class RackspaceAccessRule:
    """
    An access rule allows or denies traffic to a Load Balancer based on the
    incoming IPs.

    :param id: Unique identifier to refer to this rule by.
    :type id: ``str``

    :param rule_type: RackspaceAccessRuleType.ALLOW or
                      RackspaceAccessRuleType.DENY.
    :type id: ``int``

    :param address: IP address or cidr (can be IPv4 or IPv6).
    :type address: ``str``
    """

    def __init__(self, id=None, rule_type=None, address=None):
        self.id = id
        self.rule_type = rule_type
        self.address = address

    def _to_dict(self):
        type_string = RackspaceAccessRuleType._RULE_TYPE_STRING_MAP[self.rule_type]

        as_dict = {"type": type_string, "address": self.address}

        if self.id is not None:
            as_dict["id"] = self.id

        return as_dict


class RackspaceConnection(RackspaceConnection, PollingConnection):
    responseCls = RackspaceResponse
    auth_url = AUTH_URL
    poll_interval = 2
    timeout = 80
    cache_busting = True

    def request(self, action, params=None, data="", headers=None, method="GET"):
        if not headers:
            headers = {}
        if not params:
            params = {}

        if method in ("POST", "PUT"):
            headers["Content-Type"] = "application/json"

        return super().request(
            action=action, params=params, data=data, method=method, headers=headers
        )

    def get_poll_request_kwargs(self, response, context, request_kwargs):
        return {"action": request_kwargs["action"], "method": "GET"}

    def has_completed(self, response):
        state = response.object["loadBalancer"]["status"]
        if state == "ERROR":
            raise LibcloudError("Load balancer entered an ERROR state.", driver=self.driver)

        return state == "ACTIVE"

    def encode_data(self, data):
        return data


class RackspaceLBDriver(Driver, OpenStackDriverMixin):
    connectionCls = RackspaceConnection
    api_name = "rackspace_lb"
    name = "Rackspace LB"
    website = "http://www.rackspace.com/"

    LB_STATE_MAP = {
        "ACTIVE": State.RUNNING,
        "BUILD": State.PENDING,
        "ERROR": State.ERROR,
        "DELETED": State.DELETED,
        "PENDING_UPDATE": State.PENDING,
        "PENDING_DELETE": State.PENDING,
    }

    LB_MEMBER_CONDITION_MAP = {
        "ENABLED": MemberCondition.ENABLED,
        "DISABLED": MemberCondition.DISABLED,
        "DRAINING": MemberCondition.DRAINING,
    }

    CONDITION_LB_MEMBER_MAP = reverse_dict(LB_MEMBER_CONDITION_MAP)

    _VALUE_TO_ALGORITHM_MAP = {
        "RANDOM": Algorithm.RANDOM,
        "ROUND_ROBIN": Algorithm.ROUND_ROBIN,
        "LEAST_CONNECTIONS": Algorithm.LEAST_CONNECTIONS,
        "WEIGHTED_ROUND_ROBIN": Algorithm.WEIGHTED_ROUND_ROBIN,
        "WEIGHTED_LEAST_CONNECTIONS": Algorithm.WEIGHTED_LEAST_CONNECTIONS,
    }

    _ALGORITHM_TO_VALUE_MAP = reverse_dict(_VALUE_TO_ALGORITHM_MAP)

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        region="ord",
        **kwargs,
    ):
        ex_force_region = kwargs.pop("ex_force_region", None)
        if ex_force_region:
            # For backward compatibility
            region = ex_force_region
        OpenStackDriverMixin.__init__(self, **kwargs)
        super().__init__(key=key, secret=secret, secure=secure, host=host, port=port, region=region)

    @classmethod
    def list_regions(cls):
        return ENDPOINT_ARGS_MAP.keys()

    def _ex_connection_class_kwargs(self):
        endpoint_args = ENDPOINT_ARGS_MAP[self.region]
        kwargs = self.openstack_connection_kwargs()
        kwargs["get_endpoint_args"] = endpoint_args
        return kwargs

    def list_protocols(self):
        return self._to_protocols(self.connection.request("/loadbalancers/protocols").object)

    def ex_list_protocols_with_default_ports(self):
        """
        List protocols with default ports.

        :rtype: ``list`` of ``tuple``
        :return: A list of protocols with default ports included.
        """
        return self._to_protocols_with_default_ports(
            self.connection.request("/loadbalancers/protocols").object
        )

    def list_balancers(
        self,
        ex_member_address=None,
        ex_status=None,
        ex_changes_since=None,
        ex_params={},
    ):
        """
        @inherits: :class:`Driver.list_balancers`

        :param ex_member_address: Optional IP address of the attachment member.
                                  If provided, only the load balancers which
                                  have this member attached will be returned.
        :type ex_member_address: ``str``

        :param ex_status: Optional. Filter balancers by status
        :type ex_status: ``str``

        :param ex_changes_since: Optional. List all load balancers that have
                                 changed since the specified date/time
        :type ex_changes_since: ``str``

        :param ex_params: Optional. Set parameters to be submitted to the API
                          in the query string
        :type ex_params: ``dict``
        """

        params = {}

        if ex_member_address:
            params["nodeaddress"] = ex_member_address

        if ex_status:
            params["status"] = ex_status

        if ex_changes_since:
            params["changes-since"] = ex_changes_since

        for key, value in ex_params.items():
            params[key] = value

        return self._to_balancers(self.connection.request("/loadbalancers", params=params).object)

    def create_balancer(self, name, members, protocol="http", port=80, algorithm=DEFAULT_ALGORITHM):
        return self.ex_create_balancer(name, members, protocol, port, algorithm)

    def ex_create_balancer(
        self,
        name,
        members,
        protocol="http",
        port=80,
        algorithm=DEFAULT_ALGORITHM,
        vip="PUBLIC",
    ):
        """
        Creates a new load balancer instance

        :param name: Name of the new load balancer (required)
        :type  name: ``str``

        :param members: ``list`` of:class:`Member`s to attach to balancer
        :type  members: ``list`` of :class:`Member`

        :param protocol: Loadbalancer protocol, defaults to http.
        :type  protocol: ``str``

        :param port: Port the load balancer should listen on, defaults to 80
        :type  port: ``str``

        :param algorithm: Load balancing algorithm, defaults to
                            LBAlgorithm.ROUND_ROBIN
        :type  algorithm: :class:`Algorithm`

        :param vip: Virtual ip type of PUBLIC, SERVICENET, or ID of a virtual
                      ip
        :type  vip: ``str``

        :rtype: :class:`LoadBalancer`
        """
        balancer_attrs = self._kwargs_to_mutable_attrs(
            name=name, protocol=protocol, port=port, algorithm=algorithm, vip=vip
        )

        balancer_attrs.update({"nodes": [self._member_attributes(member) for member in members]})
        # balancer_attrs['nodes'] = ['fu']
        balancer_object = {"loadBalancer": balancer_attrs}

        resp = self.connection.request(
            "/loadbalancers", method="POST", data=json.dumps(balancer_object)
        )
        return self._to_balancer(resp.object["loadBalancer"])

    def _member_attributes(self, member):
        member_attributes = {"address": member.ip, "port": member.port}

        member_attributes.update(self._kwargs_to_mutable_member_attrs(**member.extra))

        # If the condition is not specified on the member, then it should be
        # set to ENABLED by default
        if "condition" not in member_attributes:
            member_attributes["condition"] = self.CONDITION_LB_MEMBER_MAP[MemberCondition.ENABLED]

        return member_attributes

    def destroy_balancer(self, balancer):
        uri = "/loadbalancers/%s" % (balancer.id)
        resp = self.connection.request(uri, method="DELETE")

        return resp.status == httplib.ACCEPTED

    def ex_destroy_balancers(self, balancers):
        """
        Destroys a list of Balancers (the API supports up to 10).

        :param balancers: A list of Balancers to destroy.
        :type balancers: ``list`` of :class:`LoadBalancer`

        :return: Returns whether the destroy request was accepted.
        :rtype: ``bool``
        """
        ids = [("id", balancer.id) for balancer in balancers]
        resp = self.connection.request("/loadbalancers", method="DELETE", params=ids)

        return resp.status == httplib.ACCEPTED

    def get_balancer(self, balancer_id):
        uri = "/loadbalancers/%s" % (balancer_id)
        resp = self.connection.request(uri)

        return self._to_balancer(resp.object["loadBalancer"])

    def balancer_attach_member(self, balancer, member):
        member_object = {"nodes": [self._member_attributes(member)]}

        uri = "/loadbalancers/%s/nodes" % (balancer.id)
        resp = self.connection.request(uri, method="POST", data=json.dumps(member_object))
        return self._to_members(resp.object, balancer)[0]

    def ex_balancer_attach_members(self, balancer, members):
        """
        Attaches a list of members to a load balancer.

        :param balancer: The Balancer to which members will be attached.
        :type  balancer: :class:`LoadBalancer`

        :param members: A list of Members to attach.
        :type  members: ``list`` of :class:`Member`

        :rtype: ``list`` of :class:`Member`
        """
        member_objects = {"nodes": [self._member_attributes(member) for member in members]}

        uri = "/loadbalancers/%s/nodes" % (balancer.id)
        resp = self.connection.request(uri, method="POST", data=json.dumps(member_objects))
        return self._to_members(resp.object, balancer)

    def balancer_detach_member(self, balancer, member):
        # Loadbalancer always needs to have at least 1 member.
        # Last member cannot be detached. You can only disable it or destroy
        # the balancer.
        uri = "/loadbalancers/{}/nodes/{}".format(balancer.id, member.id)
        resp = self.connection.request(uri, method="DELETE")

        return resp.status == httplib.ACCEPTED

    def ex_balancer_detach_members(self, balancer, members):
        """
        Detaches a list of members from a balancer (the API supports up to 10).
        This method blocks until the detach request has been processed and the
        balancer is in a RUNNING state again.

        :param balancer: The Balancer to detach members from.
        :type  balancer: :class:`LoadBalancer`

        :param members: A list of Members to detach.
        :type  members: ``list`` of :class:`Member`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        accepted = self.ex_balancer_detach_members_no_poll(balancer, members)

        if not accepted:
            msg = "Detach members request was not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_balancer_detach_members_no_poll(self, balancer, members):
        """
        Detaches a list of members from a balancer (the API supports up to 10).
        This method returns immediately.

        :param balancer: The Balancer to detach members from.
        :type  balancer: :class:`LoadBalancer`

        :param members: A list of Members to detach.
        :type  members: ``list`` of :class:`Member`

        :return: Returns whether the detach request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/nodes" % (balancer.id)
        ids = [("id", member.id) for member in members]
        resp = self.connection.request(uri, method="DELETE", params=ids)

        return resp.status == httplib.ACCEPTED

    def balancer_list_members(self, balancer):
        uri = "/loadbalancers/%s/nodes" % (balancer.id)
        data = self.connection.request(uri).object
        return self._to_members(data, balancer)

    def update_balancer(self, balancer, **kwargs):
        attrs = self._kwargs_to_mutable_attrs(**kwargs)
        resp = self.connection.async_request(
            action="/loadbalancers/%s" % balancer.id,
            method="PUT",
            data=json.dumps(attrs),
        )
        return self._to_balancer(resp.object["loadBalancer"])

    def ex_update_balancer_no_poll(self, balancer, **kwargs):
        """
        Update balancer no poll.

        @inherits: :class:`Driver.update_balancer`
        """
        attrs = self._kwargs_to_mutable_attrs(**kwargs)
        resp = self.connection.request(
            action="/loadbalancers/%s" % balancer.id,
            method="PUT",
            data=json.dumps(attrs),
        )
        return resp.status == httplib.ACCEPTED

    def ex_balancer_update_member(self, balancer, member, **kwargs):
        """
        Updates a Member's extra attributes for a Balancer.  The attributes can
        include 'weight' or 'condition'.  This method blocks until the update
        request has been processed and the balancer is in a RUNNING state
        again.

        :param balancer: Balancer to update the member on.
        :type  balancer: :class:`LoadBalancer`

        :param member: Member which should be used
        :type member: :class:`Member`

        :keyword **kwargs: New attributes.  Should contain either 'weight'
        or 'condition'.  'condition' can be set to 'ENABLED', 'DISABLED'.
        or 'DRAINING'.  'weight' can be set to a positive integer between
        1 and 100, with a higher weight indicating that the node will receive
        more traffic (assuming the Balancer is using a weighted algorithm).
        :type **kwargs: ``dict``

        :return: Updated Member.
        :rtype: :class:`Member`
        """
        accepted = self.ex_balancer_update_member_no_poll(balancer, member, **kwargs)

        if not accepted:
            msg = "Update member attributes was not accepted"
            raise LibcloudError(msg, driver=self)

        balancer = self._get_updated_balancer(balancer)
        members = balancer.extra["members"]

        updated_members = [m for m in members if m.id == member.id]

        if not updated_members:
            raise LibcloudError("Could not find updated member")

        return updated_members[0]

    def ex_balancer_update_member_no_poll(self, balancer, member, **kwargs):
        """
        Updates a Member's extra attributes for a Balancer.  The attribute can
        include 'weight' or 'condition'.  This method returns immediately.

        :param balancer: Balancer to update the member on.
        :type balancer: :class:`LoadBalancer`

        :param member: Member which should be used
        :type member: :class:`Member`

        :keyword **kwargs: New attributes.  Should contain either 'weight'
        or 'condition'.  'condition' can be set to 'ENABLED', 'DISABLED'.
        or 'DRAINING'.  'weight' can be set to a positive integer between
        1 and 100, with a higher weight indicating that the node will receive
        more traffic (assuming the Balancer is using a weighted algorithm).
        :type **kwargs: ``dict``

        :return: Returns whether the update request was accepted.
        :rtype: ``bool``
        """
        resp = self.connection.request(
            action="/loadbalancers/{}/nodes/{}".format(balancer.id, member.id),
            method="PUT",
            data=json.dumps(self._kwargs_to_mutable_member_attrs(**kwargs)),
        )

        return resp.status == httplib.ACCEPTED

    def ex_list_algorithm_names(self):
        """
        Lists algorithms supported by the API.  Returned as strings because
        this list may change in the future.

        :rtype: ``list`` of ``str``
        """
        response = self.connection.request("/loadbalancers/algorithms")
        return [a["name"].upper() for a in response.object["algorithms"]]

    def ex_get_balancer_error_page(self, balancer):
        """
        List error page configured for the specified load balancer.

        :param balancer: Balancer which should be used
        :type balancer: :class:`LoadBalancer`

        :rtype: ``str``
        """
        uri = "/loadbalancers/%s/errorpage" % (balancer.id)
        resp = self.connection.request(uri)

        return resp.object["errorpage"]["content"]

    def ex_balancer_access_list(self, balancer):
        """
        List the access list.

        :param balancer: Balancer which should be used
        :type balancer: :class:`LoadBalancer`

        :rtype: ``list`` of :class:`RackspaceAccessRule`
        """
        uri = "/loadbalancers/%s/accesslist" % (balancer.id)
        resp = self.connection.request(uri)

        return [self._to_access_rule(el) for el in resp.object["accessList"]]

    def _get_updated_balancer(self, balancer):
        """
        Updating a balancer's attributes puts a balancer into
        'PENDING_UPDATE' status.  Wait until the balancer is
        back in 'ACTIVE' status and then return the individual
        balancer details call.
        """
        resp = self.connection.async_request(action="/loadbalancers/%s" % balancer.id, method="GET")

        return self._to_balancer(resp.object["loadBalancer"])

    def ex_update_balancer_health_monitor(self, balancer, health_monitor):
        """
        Sets a Balancer's health monitor.  This method blocks until the update
        request has been processed and the balancer is in a RUNNING state
        again.

        :param balancer: Balancer to update.
        :type  balancer: :class:`LoadBalancer`

        :param health_monitor: Health Monitor for the balancer.
        :type  health_monitor: :class:`RackspaceHealthMonitor`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        accepted = self.ex_update_balancer_health_monitor_no_poll(balancer, health_monitor)
        if not accepted:
            msg = "Update health monitor request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_update_balancer_health_monitor_no_poll(self, balancer, health_monitor):
        """
        Sets a Balancer's health monitor.  This method returns immediately.

        :param balancer: Balancer to update health monitor on.
        :type  balancer: :class:`LoadBalancer`

        :param health_monitor: Health Monitor for the balancer.
        :type  health_monitor: :class:`RackspaceHealthMonitor`

        :return: Returns whether the update request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/healthmonitor" % (balancer.id)

        resp = self.connection.request(
            uri, method="PUT", data=json.dumps(health_monitor._to_dict())
        )

        return resp.status == httplib.ACCEPTED

    def ex_disable_balancer_health_monitor(self, balancer):
        """
        Disables a Balancer's health monitor.  This method blocks until the
        disable request has been processed and the balancer is in a RUNNING
        state again.

        :param balancer: Balancer to disable health monitor on.
        :type  balancer: :class:`LoadBalancer`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        if not self.ex_disable_balancer_health_monitor_no_poll(balancer):
            msg = "Disable health monitor request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_disable_balancer_health_monitor_no_poll(self, balancer):
        """
        Disables a Balancer's health monitor.  This method returns
        immediately.

        :param balancer: Balancer to disable health monitor on.
        :type  balancer: :class:`LoadBalancer`

        :return: Returns whether the disable request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/healthmonitor" % (balancer.id)

        resp = self.connection.request(uri, method="DELETE")

        return resp.status == httplib.ACCEPTED

    def ex_update_balancer_connection_throttle(self, balancer, connection_throttle):
        """
        Updates a Balancer's connection throttle.  This method blocks until
        the update request has been processed and the balancer is in a
        RUNNING state again.

        :param balancer: Balancer to update connection throttle on.
        :type  balancer: :class:`LoadBalancer`

        :param connection_throttle: Connection Throttle for the balancer.
        :type  connection_throttle: :class:`RackspaceConnectionThrottle`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        accepted = self.ex_update_balancer_connection_throttle_no_poll(
            balancer, connection_throttle
        )

        if not accepted:
            msg = "Update connection throttle request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_update_balancer_connection_throttle_no_poll(self, balancer, connection_throttle):
        """
        Sets a Balancer's connection throttle.  This method returns
        immediately.

        :param balancer: Balancer to update connection throttle on.
        :type  balancer: :class:`LoadBalancer`

        :param connection_throttle: Connection Throttle for the balancer.
        :type  connection_throttle: :class:`RackspaceConnectionThrottle`

        :return: Returns whether the update request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/connectionthrottle" % (balancer.id)
        resp = self.connection.request(
            uri, method="PUT", data=json.dumps(connection_throttle._to_dict())
        )

        return resp.status == httplib.ACCEPTED

    def ex_disable_balancer_connection_throttle(self, balancer):
        """
        Disables a Balancer's connection throttle.  This method blocks until
        the disable request has been processed and the balancer is in a RUNNING
        state again.

        :param balancer: Balancer to disable connection throttle on.
        :type  balancer: :class:`LoadBalancer`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        if not self.ex_disable_balancer_connection_throttle_no_poll(balancer):
            msg = "Disable connection throttle request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_disable_balancer_connection_throttle_no_poll(self, balancer):
        """
        Disables a Balancer's connection throttle.  This method returns
        immediately.

        :param balancer: Balancer to disable connection throttle on.
        :type  balancer: :class:`LoadBalancer`

        :return: Returns whether the disable request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/connectionthrottle" % (balancer.id)
        resp = self.connection.request(uri, method="DELETE")

        return resp.status == httplib.ACCEPTED

    def ex_enable_balancer_connection_logging(self, balancer):
        """
        Enables connection logging for a Balancer.  This method blocks until
        the enable request has been processed and the balancer is in a RUNNING
        state again.

        :param balancer: Balancer to enable connection logging on.
        :type  balancer: :class:`LoadBalancer`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        if not self.ex_enable_balancer_connection_logging_no_poll(balancer):
            msg = "Enable connection logging request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_enable_balancer_connection_logging_no_poll(self, balancer):
        """
        Enables connection logging for a Balancer.  This method returns
        immediately.

        :param balancer: Balancer to enable connection logging on.
        :type  balancer: :class:`LoadBalancer`

        :return: Returns whether the enable request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/connectionlogging" % (balancer.id)

        resp = self.connection.request(
            uri, method="PUT", data=json.dumps({"connectionLogging": {"enabled": True}})
        )

        return resp.status == httplib.ACCEPTED

    def ex_disable_balancer_connection_logging(self, balancer):
        """
        Disables connection logging for a Balancer.  This method blocks until
        the enable request has been processed and the balancer is in a RUNNING
        state again.

        :param balancer: Balancer to disable connection logging on.
        :type  balancer: :class:`LoadBalancer`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        if not self.ex_disable_balancer_connection_logging_no_poll(balancer):
            msg = "Disable connection logging request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_disable_balancer_connection_logging_no_poll(self, balancer):
        """
        Disables connection logging for a Balancer.  This method returns
        immediately.

        :param balancer: Balancer to disable connection logging on.
        :type  balancer: :class:`LoadBalancer`

        :return: Returns whether the disable request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/connectionlogging" % (balancer.id)
        resp = self.connection.request(
            uri,
            method="PUT",
            data=json.dumps({"connectionLogging": {"enabled": False}}),
        )

        return resp.status == httplib.ACCEPTED

    def ex_enable_balancer_session_persistence(self, balancer):
        """
        Enables session persistence for a Balancer by setting the persistence
        type to 'HTTP_COOKIE'.  This method blocks until the enable request
        has been processed and the balancer is in a RUNNING state again.

        :param balancer: Balancer to enable session persistence on.
        :type  balancer: :class:`LoadBalancer`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        if not self.ex_enable_balancer_session_persistence_no_poll(balancer):
            msg = "Enable session persistence request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_enable_balancer_session_persistence_no_poll(self, balancer):
        """
        Enables session persistence for a Balancer by setting the persistence
        type to 'HTTP_COOKIE'.  This method returns immediately.

        :param balancer: Balancer to enable session persistence on.
        :type  balancer: :class:`LoadBalancer`

        :return: Returns whether the enable request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/sessionpersistence" % (balancer.id)
        resp = self.connection.request(
            uri,
            method="PUT",
            data=json.dumps({"sessionPersistence": {"persistenceType": "HTTP_COOKIE"}}),
        )

        return resp.status == httplib.ACCEPTED

    def ex_disable_balancer_session_persistence(self, balancer):
        """
        Disables session persistence for a Balancer.  This method blocks until
        the disable request has been processed and the balancer is in a RUNNING
        state again.

        :param balancer: Balancer to disable session persistence on.
        :type balancer:  :class:`LoadBalancer`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        if not self.ex_disable_balancer_session_persistence_no_poll(balancer):
            msg = "Disable session persistence request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_disable_balancer_session_persistence_no_poll(self, balancer):
        """
        Disables session persistence for a Balancer.  This method returns
        immediately.

        :param balancer: Balancer to disable session persistence for.
        :type  balancer: :class:`LoadBalancer`

        :return: Returns whether the disable request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/sessionpersistence" % (balancer.id)
        resp = self.connection.request(uri, method="DELETE")

        return resp.status == httplib.ACCEPTED

    def ex_update_balancer_error_page(self, balancer, page_content):
        """
        Updates a Balancer's custom error page.  This method blocks until
        the update request has been processed and the balancer is in a
        RUNNING state again.

        :param balancer: Balancer to update the custom error page for.
        :type  balancer: :class:`LoadBalancer`

        :param page_content: HTML content for the custom error page.
        :type  page_content: ``str``

        :return: Updated Balancer.
        :rtype:  :class:`LoadBalancer`
        """
        accepted = self.ex_update_balancer_error_page_no_poll(balancer, page_content)
        if not accepted:
            msg = "Update error page request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_update_balancer_error_page_no_poll(self, balancer, page_content):
        """
        Updates a Balancer's custom error page.  This method returns
        immediately.

        :param balancer: Balancer to update the custom error page for.
        :type  balancer: :class:`LoadBalancer`

        :param page_content: HTML content for the custom error page.
        :type  page_content: ``str``

        :return: Returns whether the update request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/errorpage" % (balancer.id)
        resp = self.connection.request(
            uri, method="PUT", data=json.dumps({"errorpage": {"content": page_content}})
        )

        return resp.status == httplib.ACCEPTED

    def ex_disable_balancer_custom_error_page(self, balancer):
        """
        Disables a Balancer's custom error page, returning its error page to
        the Rackspace-provided default.  This method blocks until the disable
        request has been processed and the balancer is in a RUNNING state
        again.

        :param balancer: Balancer to disable the custom error page for.
        :type  balancer: :class:`LoadBalancer`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        if not self.ex_disable_balancer_custom_error_page_no_poll(balancer):
            msg = "Disable custom error page request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_disable_balancer_custom_error_page_no_poll(self, balancer):
        """
        Disables a Balancer's custom error page, returning its error page to
        the Rackspace-provided default.  This method returns immediately.

        :param balancer: Balancer to disable the custom error page for.
        :type  balancer: :class:`LoadBalancer`

        :return: Returns whether the disable request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/errorpage" % (balancer.id)
        resp = self.connection.request(uri, method="DELETE")

        # Load Balancer API currently returns 200 OK on custom error page
        # delete.
        return resp.status == httplib.OK or resp.status == httplib.ACCEPTED

    def ex_create_balancer_access_rule(self, balancer, rule):
        """
        Adds an access rule to a Balancer's access list.  This method blocks
        until the update request has been processed and the balancer is in a
        RUNNING state again.

        :param balancer: Balancer to create the access rule for.
        :type balancer: :class:`LoadBalancer`

        :param rule: Access Rule to add to the balancer.
        :type rule: :class:`RackspaceAccessRule`

        :return: The created access rule.
        :rtype: :class:`RackspaceAccessRule`
        """
        accepted = self.ex_create_balancer_access_rule_no_poll(balancer, rule)
        if not accepted:
            msg = "Create access rule not accepted"
            raise LibcloudError(msg, driver=self)

        balancer = self._get_updated_balancer(balancer)
        access_list = balancer.extra["accessList"]

        created_rule = self._find_matching_rule(rule, access_list)

        if not created_rule:
            raise LibcloudError("Could not find created rule")

        return created_rule

    def ex_create_balancer_access_rule_no_poll(self, balancer, rule):
        """
        Adds an access rule to a Balancer's access list.  This method returns
        immediately.

        :param balancer: Balancer to create the access rule for.
        :type balancer: :class:`LoadBalancer`

        :param rule: Access Rule to add to the balancer.
        :type rule: :class:`RackspaceAccessRule`

        :return: Returns whether the create request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/accesslist" % (balancer.id)
        resp = self.connection.request(
            uri, method="POST", data=json.dumps({"networkItem": rule._to_dict()})
        )

        return resp.status == httplib.ACCEPTED

    def ex_create_balancer_access_rules(self, balancer, rules):
        """
        Adds a list of access rules to a Balancer's access list.  This method
        blocks until the update request has been processed and the balancer is
        in a RUNNING state again.

        :param balancer: Balancer to create the access rule for.
        :type  balancer: :class:`LoadBalancer`

        :param rules: List of :class:`RackspaceAccessRule` to add to the
                      balancer.
        :type  rules: ``list`` of :class:`RackspaceAccessRule`

        :return: The created access rules.
        :rtype: :class:`RackspaceAccessRule`
        """
        accepted = self.ex_create_balancer_access_rules_no_poll(balancer, rules)
        if not accepted:
            msg = "Create access rules not accepted"
            raise LibcloudError(msg, driver=self)

        balancer = self._get_updated_balancer(balancer)
        access_list = balancer.extra["accessList"]

        created_rules = []
        for r in rules:
            matched_rule = self._find_matching_rule(r, access_list)
            if matched_rule:
                created_rules.append(matched_rule)

        if len(created_rules) != len(rules):
            raise LibcloudError("Could not find all created rules")

        return created_rules

    def _find_matching_rule(self, rule_to_find, access_list):
        """
        LB API does not return the ID for the newly created rules, so we have
        to search the list to find the rule with a matching rule type and
        address to return an object with the right identifier.it.  The API
        enforces rule type and address uniqueness.
        """
        for r in access_list:
            if rule_to_find.rule_type == r.rule_type and rule_to_find.address == r.address:
                return r

        return None

    def ex_create_balancer_access_rules_no_poll(self, balancer, rules):
        """
        Adds a list of access rules to a Balancer's access list.  This method
        returns immediately.

        :param balancer: Balancer to create the access rule for.
        :type balancer: :class:`LoadBalancer`

        :param rules: List of :class:`RackspaceAccessRule` to add to
                      the balancer.
        :type  rules: ``list`` of :class:`RackspaceAccessRule`

        :return: Returns whether the create request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/%s/accesslist" % (balancer.id)
        resp = self.connection.request(
            uri,
            method="POST",
            data=json.dumps({"accessList": [rule._to_dict() for rule in rules]}),
        )

        return resp.status == httplib.ACCEPTED

    def ex_destroy_balancer_access_rule(self, balancer, rule):
        """
        Removes an access rule from a Balancer's access list.  This method
        blocks until the update request has been processed and the balancer
        is in a RUNNING state again.

        :param balancer: Balancer to remove the access rule from.
        :type  balancer: :class:`LoadBalancer`

        :param rule: Access Rule to remove from the balancer.
        :type  rule: :class:`RackspaceAccessRule`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        accepted = self.ex_destroy_balancer_access_rule_no_poll(balancer, rule)
        if not accepted:
            msg = "Delete access rule not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_destroy_balancer_access_rule_no_poll(self, balancer, rule):
        """
        Removes an access rule from a Balancer's access list.  This method
        returns immediately.

        :param balancer: Balancer to remove the access rule from.
        :type  balancer: :class:`LoadBalancer`

        :param rule: Access Rule to remove from the balancer.
        :type  rule: :class:`RackspaceAccessRule`

        :return: Returns whether the destroy request was accepted.
        :rtype: ``bool``
        """
        uri = "/loadbalancers/{}/accesslist/{}".format(balancer.id, rule.id)
        resp = self.connection.request(uri, method="DELETE")

        return resp.status == httplib.ACCEPTED

    def ex_destroy_balancer_access_rules(self, balancer, rules):
        """
        Removes a list of access rules from a Balancer's access list.  This
        method blocks until the update request has been processed and the
        balancer is in a RUNNING state again.

        :param balancer: Balancer to remove the access rules from.
        :type  balancer: :class:`LoadBalancer`

        :param rules: List of :class:`RackspaceAccessRule` objects to remove
                      from the balancer.
        :type  rules: ``list`` of :class:`RackspaceAccessRule`

        :return: Updated Balancer.
        :rtype: :class:`LoadBalancer`
        """
        accepted = self.ex_destroy_balancer_access_rules_no_poll(balancer, rules)

        if not accepted:
            msg = "Destroy access rules request not accepted"
            raise LibcloudError(msg, driver=self)

        return self._get_updated_balancer(balancer)

    def ex_destroy_balancer_access_rules_no_poll(self, balancer, rules):
        """
        Removes a list of access rules from a Balancer's access list.  This
        method returns immediately.

        :param balancer: Balancer to remove the access rules from.
        :type  balancer: :class:`LoadBalancer`

        :param rules: List of :class:`RackspaceAccessRule` objects to remove
                      from the balancer.
        :type  rules: ``list`` of :class:`RackspaceAccessRule`

        :return: Returns whether the destroy request was accepted.
        :rtype: ``bool``
        """
        ids = [("id", rule.id) for rule in rules]
        uri = "/loadbalancers/%s/accesslist" % balancer.id

        resp = self.connection.request(uri, method="DELETE", params=ids)

        return resp.status == httplib.ACCEPTED

    def ex_list_current_usage(self, balancer):
        """
        Return current load balancer usage report.

        :param balancer: Balancer to remove the access rules from.
        :type  balancer: :class:`LoadBalancer`

        :return: Raw load balancer usage object.
        :rtype: ``dict``
        """
        uri = "/loadbalancers/%s/usage/current" % (balancer.id)
        resp = self.connection.request(uri, method="GET")

        return resp.object

    def _to_protocols(self, object):
        protocols = []
        for item in object["protocols"]:
            protocols.append(item["name"].lower())
        return protocols

    def _to_protocols_with_default_ports(self, object):
        protocols = []
        for item in object["protocols"]:
            name = item["name"].lower()
            port = int(item["port"])
            protocols.append((name, port))

        return protocols

    def _to_balancers(self, object):
        return [self._to_balancer(el) for el in object["loadBalancers"]]

    def _to_balancer(self, el):
        ip = None
        port = None
        sourceAddresses = {}

        if "port" in el:
            port = el["port"]

        if "sourceAddresses" in el:
            sourceAddresses = el["sourceAddresses"]

        extra = {
            "ipv6PublicSource": sourceAddresses.get("ipv6Public"),
            "ipv4PublicSource": sourceAddresses.get("ipv4Public"),
            "ipv4PrivateSource": sourceAddresses.get("ipv4Servicenet"),
            "service_name": self.connection.get_service_name(),
            "uri": "https://%s%s/loadbalancers/%s"
            % (self.connection.host, self.connection.request_path, el["id"]),
        }

        if "virtualIps" in el:
            ip = el["virtualIps"][0]["address"]
            extra["virtualIps"] = el["virtualIps"]

        if "protocol" in el:
            extra["protocol"] = el["protocol"]

        if "algorithm" in el and el["algorithm"] in self._VALUE_TO_ALGORITHM_MAP:
            extra["algorithm"] = self._value_to_algorithm(el["algorithm"])

        if "healthMonitor" in el:
            health_monitor = self._to_health_monitor(el)
            if health_monitor:
                extra["healthMonitor"] = health_monitor

        if "connectionThrottle" in el:
            extra["connectionThrottle"] = self._to_connection_throttle(el)

        if "sessionPersistence" in el:
            persistence = el["sessionPersistence"]
            extra["sessionPersistenceType"] = persistence.get("persistenceType")

        if "connectionLogging" in el:
            logging = el["connectionLogging"]
            extra["connectionLoggingEnabled"] = logging.get("enabled")

        if "nodes" in el:
            extra["members"] = self._to_members(el)

        if "created" in el:
            extra["created"] = self._iso_to_datetime(el["created"]["time"])

        if "updated" in el:
            extra["updated"] = self._iso_to_datetime(el["updated"]["time"])

        if "accessList" in el:
            extra["accessList"] = [self._to_access_rule(rule) for rule in el["accessList"]]

        return LoadBalancer(
            id=el["id"],
            name=el["name"],
            state=self.LB_STATE_MAP.get(el["status"], State.UNKNOWN),
            ip=ip,
            port=port,
            driver=self.connection.driver,
            extra=extra,
        )

    def _to_members(self, object, balancer=None):
        return [self._to_member(el, balancer) for el in object["nodes"]]

    def _to_member(self, el, balancer=None):
        extra = {}
        if "weight" in el:
            extra["weight"] = el["weight"]

        if "condition" in el and el["condition"] in self.LB_MEMBER_CONDITION_MAP:
            extra["condition"] = self.LB_MEMBER_CONDITION_MAP.get(el["condition"])

        if "status" in el:
            extra["status"] = el["status"]

        lbmember = Member(
            id=el["id"],
            ip=el["address"],
            port=el["port"],
            balancer=balancer,
            extra=extra,
        )
        return lbmember

    def _protocol_to_value(self, protocol):
        non_standard_protocols = {
            "imapv2": "IMAPv2",
            "imapv3": "IMAPv3",
            "imapv4": "IMAPv4",
        }
        protocol_name = protocol.lower()

        if protocol_name in non_standard_protocols:
            protocol_value = non_standard_protocols[protocol_name]
        else:
            protocol_value = protocol.upper()

        return protocol_value

    def _kwargs_to_mutable_attrs(self, **attrs):
        update_attrs = {}
        if "name" in attrs:
            update_attrs["name"] = attrs["name"]

        if "algorithm" in attrs:
            algorithm_value = self._algorithm_to_value(attrs["algorithm"])
            update_attrs["algorithm"] = algorithm_value

        if "protocol" in attrs:
            update_attrs["protocol"] = self._protocol_to_value(attrs["protocol"])

        if "port" in attrs:
            update_attrs["port"] = int(attrs["port"])

        if "vip" in attrs:
            if attrs["vip"] == "PUBLIC" or attrs["vip"] == "SERVICENET":
                update_attrs["virtualIps"] = [{"type": attrs["vip"]}]
            else:
                update_attrs["virtualIps"] = [{"id": attrs["vip"]}]

        return update_attrs

    def _kwargs_to_mutable_member_attrs(self, **attrs):
        update_attrs = {}
        if "condition" in attrs:
            update_attrs["condition"] = self.CONDITION_LB_MEMBER_MAP.get(attrs["condition"])

        if "weight" in attrs:
            update_attrs["weight"] = attrs["weight"]

        return update_attrs

    def _to_health_monitor(self, el):
        health_monitor_data = el["healthMonitor"]

        type = health_monitor_data.get("type")
        delay = health_monitor_data.get("delay")
        timeout = health_monitor_data.get("timeout")
        attempts_before_deactivation = health_monitor_data.get("attemptsBeforeDeactivation")

        if type == "CONNECT":
            return RackspaceHealthMonitor(
                type=type,
                delay=delay,
                timeout=timeout,
                attempts_before_deactivation=attempts_before_deactivation,
            )

        if type == "HTTP" or type == "HTTPS":
            return RackspaceHTTPHealthMonitor(
                type=type,
                delay=delay,
                timeout=timeout,
                attempts_before_deactivation=attempts_before_deactivation,
                path=health_monitor_data.get("path"),
                status_regex=health_monitor_data.get("statusRegex"),
                body_regex=health_monitor_data.get("bodyRegex", ""),
            )

        return None

    def _to_connection_throttle(self, el):
        connection_throttle_data = el["connectionThrottle"]

        min_connections = connection_throttle_data.get("minConnections")
        max_connections = connection_throttle_data.get("maxConnections")
        max_connection_rate = connection_throttle_data.get("maxConnectionRate")
        rate_interval = connection_throttle_data.get("rateInterval")

        return RackspaceConnectionThrottle(
            min_connections=min_connections,
            max_connections=max_connections,
            max_connection_rate=max_connection_rate,
            rate_interval_seconds=rate_interval,
        )

    def _to_access_rule(self, el):
        return RackspaceAccessRule(
            id=el.get("id"),
            rule_type=self._to_access_rule_type(el.get("type")),
            address=el.get("address"),
        )

    def _to_access_rule_type(self, type):
        if type == "ALLOW":
            return RackspaceAccessRuleType.ALLOW
        elif type == "DENY":
            return RackspaceAccessRuleType.DENY

    def _iso_to_datetime(self, isodate):
        date_formats = ("%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%dT%H:%M:%S%z")
        date = None

        for date_format in date_formats:
            try:
                date = datetime.strptime(isodate, date_format)
            except ValueError:
                pass

            if date:
                break

        return date



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import (
    Provider,
    RecordType,
    RecordError,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
    RecordAlreadyExistsError,
)
from libcloud.common.gandi_live import (
    GandiLiveResponse,
    BaseGandiLiveDriver,
    GandiLiveConnection,
    ResourceConflictError,
    ResourceNotFoundError,
)

__all__ = [
    "GandiLiveDNSDriver",
]


TTL_MIN = 300
TTL_MAX = 2592000  # 30 days
API_BASE = "/api/v5"


class GandiLiveDNSResponse(GandiLiveResponse):
    pass


class GandiLiveDNSConnection(GandiLiveConnection):
    responseCls = GandiLiveDNSResponse


class GandiLiveDNSDriver(BaseGandiLiveDriver, DNSDriver):
    """
    API reference can be found at:

    https://doc.livedns.gandi.net/

    Please note that the Libcloud paradigm of one zone per domain does not
    match exactly with Gandi LiveDNS.  For Gandi, a "zone" can apply to
    multiple domains.  This driver behaves as if the domain is a zone, but be
    warned that modifying a domain means modifying the zone.  If you have a
    zone associated with multiple domains, all of those domains will be
    modified as well.
    """

    type = Provider.GANDI
    name = "Gandi LiveDNS"
    website = "http://www.gandi.net/domain"

    connectionCls = GandiLiveDNSConnection

    # also supports CAA, CDS
    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.ALIAS: "ALIAS",
        RecordType.CNAME: "CNAME",
        RecordType.DNAME: "DNAME",
        RecordType.DS: "DS",
        RecordType.KEY: "KEY",
        RecordType.LOC: "LOC",
        RecordType.MX: "MX",
        RecordType.NS: "NS",
        RecordType.PTR: "PTR",
        RecordType.SPF: "SPF",
        RecordType.SRV: "SRV",
        RecordType.SSHFP: "SSHFP",
        RecordType.TLSA: "TLSA",
        RecordType.TXT: "TXT",
        RecordType.WKS: "WKS",
        RecordType.CAA: "CAA",
    }

    def list_zones(self):
        zones = self.connection.request(action="%s/domains" % API_BASE, method="GET")
        return self._to_zones(zones.object)

    def get_zone(self, zone_id):
        action = "{}/domains/{}".format(API_BASE, zone_id)
        try:
            zone = self.connection.request(action=action, method="GET")
        except ResourceNotFoundError:
            raise ZoneDoesNotExistError(value="", driver=self.connection.driver, zone_id=zone_id)
        return self._to_zone(zone.object)

    """
    :param extra: (optional) Extra attribute ('name'); if not provided, name
                             is based on domain.

    :return: :class:`Zone` with attribute zone_uuid set in extra ``dict``
    """

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        if extra and "name" in extra:
            zone_name = extra["name"]
        else:
            zone_name = "%s zone" % domain
        zone_data = {
            "name": zone_name,
        }

        try:
            new_zone = self.connection.request(
                action="%s/zones" % API_BASE, method="POST", data=zone_data
            )
        except ResourceConflictError:
            raise ZoneAlreadyExistsError(value="", driver=self.connection.driver, zone_id=zone_name)
        new_zone_uuid = new_zone.headers["location"].split("/")[-1]

        self.ex_switch_domain_gandi_zone(domain, new_zone_uuid)

        return self._to_zone({"fqdn": domain, "zone_uuid": new_zone_uuid})

    def list_records(self, zone):
        action = "{}/domains/{}/records".format(API_BASE, zone.id)
        records = self.connection.request(action=action, method="GET")
        return self._to_records(records.object, zone)

    """
    :return: :class:`Record` with the extra ``dict`` containing attribute
             other_values ``list`` of ``str`` for other values; the first
             value is returned through Record.data.
    """

    def get_record(self, zone_id, record_id):
        record_type, name = record_id.split(":", 1)
        action = "{}/domains/{}/records/{}/{}".format(API_BASE, zone_id, name, record_type)
        try:
            record = self.connection.request(action=action, method="GET")
        except ResourceNotFoundError:
            raise RecordDoesNotExistError(
                value="", driver=self.connection.driver, record_id=record_id
            )
        return self._to_record(record.object, self.get_zone(zone_id))[0]

    def create_record(self, name, zone, type, data, extra=None):
        self._validate_record(None, name, type, data, extra)

        action = "{}/domains/{}/records".format(API_BASE, zone.id)

        if type == "MX":
            data = "{} {}".format(extra["priority"], data)

        record_data = {
            "rrset_name": name,
            "rrset_type": self.RECORD_TYPE_MAP[type],
            "rrset_values": [data],
        }

        if extra is not None and "ttl" in extra:
            record_data["rrset_ttl"] = extra["ttl"]

        try:
            self.connection.request(action=action, method="POST", data=record_data)
        except ResourceConflictError:
            raise RecordAlreadyExistsError(
                value="",
                driver=self.connection.driver,
                record_id="{}:{}".format(self.RECORD_TYPE_MAP[type], name),
            )

        return self._to_record_sub(record_data, zone, data)

    """
    Ignores name and type, not allowed in an update call to the service.

    The Gandi service requires all values for a record when doing an update.
    Not providing all values during an update means the service will interpret
    it as replacing all values with the one data value.  The easiest way to
    accomplish this is to make sure the value of a get_record is used as the
    value of the record parameter.

    This method will change the value when only one exists.  When more than
    one exists, it will combine the data parameter value with the extra dict
    values contained in the list extra['_other_records'].  This method should
    only be used to make single value updates.

    To change the number of values in the value set or to change several at
    once, delete and recreate, potentially using ex_create_multi_value_record.
    """

    def update_record(self, record, name, type, data, extra):
        self._validate_record(record.id, record.name, record.type, data, extra)

        action = "{}/domains/{}/records/{}/{}".format(
            API_BASE,
            record.zone.id,
            record.name,
            self.RECORD_TYPE_MAP[record.type],
        )

        multiple_value_record = record.extra.get("_multi_value", False)
        other_records = record.extra.get("_other_records", [])

        if record.type == RecordType.MX:
            data = "{} {}".format(extra["priority"], data)

        if multiple_value_record and len(other_records) > 0:
            rvalue = [data]
            for other_record in other_records:
                if record.type == RecordType.MX:
                    rvalue.append(
                        "{} {}".format(other_record["extra"]["priority"], other_record["data"])
                    )
                else:
                    rvalue.append(other_record["data"])
        else:
            rvalue = [data]

        record_data = {"rrset_values": rvalue}

        if extra is not None and "ttl" in extra:
            record_data["rrset_ttl"] = extra["ttl"]

        try:
            self.connection.request(action=action, method="PUT", data=record_data)
        except ResourceNotFoundError:
            raise RecordDoesNotExistError(
                value="", driver=self.connection.driver, record_id=record.id
            )

        record_data["rrset_name"] = record.name
        record_data["rrset_type"] = self.RECORD_TYPE_MAP[record.type]
        return self._to_record(record_data, record.zone)[0]

    """
    The Gandi service considers all values for a name-type combination to be
    one record.  Deleting that name-type record means deleting all values for
    it.
    """

    def delete_record(self, record):
        action = "{}/domains/{}/records/{}/{}".format(
            API_BASE,
            record.zone.id,
            record.name,
            self.RECORD_TYPE_MAP[record.type],
        )
        try:
            self.connection.request(action=action, method="DELETE")
        except ResourceNotFoundError:
            raise RecordDoesNotExistError(
                value="", driver=self.connection.driver, record_id=record.id
            )
        # Originally checked for success here, but it should never reach
        # this point with anything other than HTTP 200
        return True

    def export_zone_to_bind_format(self, zone):
        action = "{}/domains/{}/records".format(API_BASE, zone.id)
        headers = {"Accept": "text/plain"}
        resp = self.connection.request(action=action, method="GET", headers=headers, raw=True)
        return resp.body

    # There is nothing you can update about a domain; you can update zones'
    # names and which zone a domain is associated with, but the domain itself
    # is basically immutable.  Instead, some ex_ methods for dealing with
    # Gandi zones.

    """
    Update the name of a Gandi zone.

    Note that a Gandi zone is not the same as a Libcloud zone.  A Gandi zone
    is a separate object type from a Gandi domain; a Gandi zone can be reused
    by multiple Gandi domains, and the actual records are associated with the
    zone directly.  This is mostly masked in this driver to make it look like
    records are associated with domains.  If you need to step out of that
    masking, use these extension methods.

    :param zone_uuid: Identifier for the Gandi zone.
    :type  zone_uuid: ``str``

    :param name: New name for the Gandi zone.
    :type  name: ``str``

    :return: ``bool``
    """

    def ex_update_gandi_zone_name(self, zone_uuid, name):
        action = "{}/zones/{}".format(API_BASE, zone_uuid)
        data = {
            "name": name,
        }
        self.connection.request(action=action, method="PATCH", data=data)
        return True

    # There is no concept of deleting domains in this API, not even to
    # disassociate a domain from a zone.  You can delete a zone, though.
    """
    Delete a Gandi zone.  This may raise a ResourceConflictError if you
    try to delete a zone that has domains still using it.

    :param zone_uuid: Identifier for the Gandi zone
    :type  zone_uuid: ``str``

    :return: ``bool``
    """

    def ex_delete_gandi_zone(self, zone_uuid):
        self.connection.request(action="{}/zones/{}".format(API_BASE, zone_uuid), method="DELETE")
        return True

    """
    Change the Gandi zone a domain is associated with.

    :param domain: Domain name to switch zones.
    :type  domain: ``str``

    :param zone_uuid: Identifier for the new Gandi zone to switch to.
    :type  zone_uuid: ``str``

    :return: ``bool``
    """

    def ex_switch_domain_gandi_zone(self, domain, zone_uuid):
        domain_data = {
            "zone_uuid": zone_uuid,
        }
        self.connection.request(
            action="{}/domains/{}".format(API_BASE, domain),
            method="PATCH",
            data=domain_data,
        )
        return True

    """
    Create a new record with multiple values.

    :param data: Record values (depends on the record type)
    :type  data: ``list`` (of ``str``)

    :return: ``list`` of :class:`Record`s
    """

    def ex_create_multi_value_record(self, name, zone, type, data, extra=None):
        self._validate_record(None, name, type, data, extra)

        action = "{}/domains/{}/records".format(API_BASE, zone.id)

        record_data = {
            "rrset_name": name,
            "rrset_type": self.RECORD_TYPE_MAP[type],
            "rrset_values": data,
        }

        if extra is not None and "ttl" in extra:
            record_data["rrset_ttl"] = extra["ttl"]

        try:
            self.connection.request(action=action, method="POST", data=record_data)
        except ResourceConflictError:
            raise RecordAlreadyExistsError(
                value="",
                driver=self.connection.driver,
                record_id="{}:{}".format(self.RECORD_TYPE_MAP[type], name),
            )
        return self._to_record(record_data, zone)

    def _to_record(self, data, zone):
        records = []
        rrset_values = data["rrset_values"]
        multiple_value_record = len(rrset_values) > 1

        for index, rrset_value in enumerate(rrset_values):
            record = self._to_record_sub(data, zone, rrset_value)
            record.extra["_multi_value"] = multiple_value_record
            if multiple_value_record:
                record.extra["_other_records"] = []
            records.append(record)

        if multiple_value_record:
            for index in range(0, len(records)):
                record = records[index]
                for other_index, other_record in enumerate(records):
                    if index == other_index:
                        continue

                    extra = copy.deepcopy(other_record.extra)
                    extra.pop("_multi_value")
                    extra.pop("_other_records")

                    item = {
                        "name": other_record.name,
                        "data": other_record.data,
                        "type": other_record.type,
                        "extra": extra,
                    }
                    record.extra["_other_records"].append(item)
        return records

    def _to_record_sub(self, data, zone, value):
        extra = {}
        ttl = data.get("rrset_ttl", None)
        if ttl is not None:
            extra["ttl"] = int(ttl)
        if data["rrset_type"] == "MX":
            priority, value = value.split()
            extra["priority"] = priority
        return Record(
            id="{}:{}".format(data["rrset_type"], data["rrset_name"]),
            name=data["rrset_name"],
            type=self._string_to_record_type(data["rrset_type"]),
            data=value,
            zone=zone,
            driver=self,
            ttl=ttl,
            extra=extra,
        )

    def _to_records(self, data, zone):
        records = []
        for r in data:
            records += self._to_record(r, zone)
        return records

    def _to_zone(self, zone):
        extra = {}
        if "zone_uuid" in zone:
            extra = {"zone_uuid": zone["zone_uuid"]}
        return Zone(
            id=str(zone["fqdn"]),
            domain=zone["fqdn"],
            type="master",
            ttl=0,
            driver=self,
            extra=extra,
        )

    def _to_zones(self, zones):
        ret = []
        for z in zones:
            ret.append(self._to_zone(z))
        return ret

    def _validate_record(self, record_id, name, record_type, data, extra):
        if len(data) > 1024:
            raise RecordError(
                "Record data must be <= 1024 characters",
                driver=self,
                record_id=record_id,
            )
        if type == "MX" or type == RecordType.MX:
            if extra is None or "priority" not in extra:
                raise RecordError(
                    "MX record must have a priority", driver=self, record_id=record_id
                )
        if extra is not None and "_other_records" in extra:
            for other_value in extra.get("_other_records", []):
                if len(other_value["data"]) > 1024:
                    raise RecordError(
                        "Record data must be <= 1024 characters",
                        driver=self,
                        record_id=record_id,
                    )
                if type == "MX" or type == RecordType.MX:
                    if other_value["extra"] is None or "priority" not in other_value["extra"]:
                        raise RecordError(
                            "MX record must have a priority",
                            driver=self,
                            record_id=record_id,
                        )
        if extra is not None and "ttl" in extra:
            if extra["ttl"] < TTL_MIN:
                raise RecordError(
                    "TTL must be at least 300 seconds", driver=self, record_id=record_id
                )
            if extra["ttl"] > TTL_MAX:
                raise RecordError("TTL must not exceed 30 days", driver=self, record_id=record_id)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.aws import SignedAWSConnection
from libcloud.common.types import LibcloudError
from libcloud.storage.drivers.s3 import API_VERSION, BaseS3Connection, BaseS3StorageDriver

__all__ = ["MinIOStorageDriver"]


class MinIOConnectionAWS4(SignedAWSConnection, BaseS3Connection):
    service_name = "s3"
    version = API_VERSION

    def __init__(
        self,
        user_id,
        key,
        secure=True,
        host=None,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        token=None,
        retry_delay=None,
        backoff=None,
        **kwargs,
    ):
        super().__init__(
            user_id,
            key,
            secure,
            host,
            port,
            url,
            timeout,
            proxy_url,
            token,
            retry_delay,
            backoff,
            4,
        )  # force aws4


class MinIOStorageDriver(BaseS3StorageDriver):
    name = "MinIO Storage Driver"
    website = "https://min.io/"
    connectionCls = MinIOConnectionAWS4
    region_name = ""

    def __init__(self, key, secret=None, secure=True, host=None, port=None):
        if host is None:
            raise LibcloudError("host argument is required", driver=self)

        self.connectionCls.host = host

        super().__init__(key=key, secret=secret, secure=secure, host=host, port=port)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
AuroraDNS DNS Driver
"""

import hmac
import json
import base64
import datetime
from hashlib import sha256

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import (
    RecordType,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
)
from libcloud.utils.py3 import b, httplib
from libcloud.common.base import JsonResponse, ConnectionUserAndKey
from libcloud.common.types import LibcloudError, ProviderError, InvalidCredsError

API_HOST = "api.auroradns.eu"

# Default TTL required by libcloud, but doesn't do anything in AuroraDNS
DEFAULT_ZONE_TTL = 3600
DEFAULT_ZONE_TYPE = "master"

VALID_RECORD_PARAMS_EXTRA = ["ttl", "prio", "health_check_id", "disabled"]


class AuroraDNSHealthCheckType:
    """
    Healthcheck type.
    """

    HTTP = "HTTP"
    HTTPS = "HTTPS"
    TCP = "TCP"


class HealthCheckError(LibcloudError):
    error_type = "HealthCheckError"

    def __init__(self, value, driver, health_check_id):
        self.health_check_id = health_check_id
        super().__init__(value=value, driver=driver)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<{} in {}, health_check_id={}, value={}>".format(
            self.error_type,
            repr(self.driver),
            self.health_check_id,
            self.value,
        )


class HealthCheckDoesNotExistError(HealthCheckError):
    error_type = "HealthCheckDoesNotExistError"


class AuroraDNSHealthCheck:
    """
    AuroraDNS Healthcheck resource.
    """

    def __init__(
        self,
        id,
        type,
        hostname,
        ipaddress,
        port,
        interval,
        path,
        threshold,
        health,
        enabled,
        zone,
        driver,
        extra=None,
    ):
        """
        :param id: Healthcheck id
        :type id: ``str``

        :param hostname: Hostname or FQDN of the target
        :type hostname: ``str``

        :param ipaddress: IPv4 or IPv6 address of the target
        :type ipaddress: ``str``

        :param port: The port on the target to monitor
        :type port: ``int``

        :param interval: The interval of the health check
        :type interval: ``int``

        :param path: The path to monitor on the target
        :type path: ``str``

        :param threshold: The threshold of before marking a check as failed
        :type threshold: ``int``

        :param health: The current health of the health check
        :type health: ``bool``

        :param enabled: If the health check is currently enabled
        :type enabled: ``bool``

        :param zone: Zone instance.
        :type zone: :class:`Zone`

        :param driver: DNSDriver instance.
        :type driver: :class:`DNSDriver`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``
        """
        self.id = str(id) if id else None
        self.type = type
        self.hostname = hostname
        self.ipaddress = ipaddress
        self.port = int(port) if port else None
        self.interval = int(interval)
        self.path = path
        self.threshold = int(threshold)
        self.health = bool(health)
        self.enabled = bool(enabled)
        self.zone = zone
        self.driver = driver
        self.extra = extra or {}

    def update(
        self,
        type=None,
        hostname=None,
        ipaddress=None,
        port=None,
        interval=None,
        path=None,
        threshold=None,
        enabled=None,
        extra=None,
    ):
        return self.driver.ex_update_healthcheck(
            healthcheck=self,
            type=type,
            hostname=hostname,
            ipaddress=ipaddress,
            port=port,
            path=path,
            interval=interval,
            threshold=threshold,
            enabled=enabled,
            extra=extra,
        )

    def delete(self):
        return self.driver.ex_delete_healthcheck(healthcheck=self)

    def __repr__(self):
        return (
            "<AuroraDNSHealthCheck: zone=%s, id=%s, type=%s, hostname=%s, "
            "ipaddress=%s, port=%d, interval=%d, health=%s, provider=%s"
            "...>"
            % (
                self.zone.id,
                self.id,
                self.type,
                self.hostname,
                self.ipaddress,
                self.port,
                self.interval,
                self.health,
                self.driver.name,
            )
        )


class AuroraDNSResponse(JsonResponse):
    def success(self):
        return self.status in [httplib.OK, httplib.CREATED, httplib.ACCEPTED]

    def parse_error(self):
        status = int(self.status)
        error = {"driver": self, "value": ""}

        if status == httplib.UNAUTHORIZED:
            error["value"] = "Authentication failed"
            raise InvalidCredsError(**error)
        elif status == httplib.FORBIDDEN:
            error["value"] = "Authorization failed"
            error["http_code"] = status
            raise ProviderError(**error)
        elif status == httplib.NOT_FOUND:
            context = self.connection.context
            if context["resource"] == "zone":
                error["zone_id"] = context["id"]
                raise ZoneDoesNotExistError(**error)
            elif context["resource"] == "record":
                error["record_id"] = context["id"]
                raise RecordDoesNotExistError(**error)
            elif context["resource"] == "healthcheck":
                error["health_check_id"] = context["id"]
                raise HealthCheckDoesNotExistError(**error)
        elif status == httplib.CONFLICT:
            context = self.connection.context
            if context["resource"] == "zone":
                error["zone_id"] = context["id"]
                raise ZoneAlreadyExistsError(**error)
        elif status == httplib.BAD_REQUEST:
            context = self.connection.context
            body = self.parse_body()
            raise ProviderError(value=body["errormsg"], http_code=status, driver=self)


class AuroraDNSConnection(ConnectionUserAndKey):
    host = API_HOST
    responseCls = AuroraDNSResponse

    def calculate_auth_signature(self, secret_key, method, url, timestamp):
        b64_hmac = base64.b64encode(
            hmac.new(b(secret_key), b(method) + b(url) + b(timestamp), digestmod=sha256).digest()
        )

        return b64_hmac.decode("utf-8")

    def gen_auth_header(self, api_key, secret_key, method, url, timestamp):
        signature = self.calculate_auth_signature(secret_key, method, url, timestamp)

        auth_b64 = base64.b64encode(b("{}:{}".format(api_key, signature)))
        return "AuroraDNSv1 %s" % (auth_b64.decode("utf-8"))

    def request(self, action, params=None, data="", headers=None, method="GET"):
        if not headers:
            headers = {}
        if not params:
            params = {}

        if method in ("POST", "PUT"):
            headers = {"Content-Type": "application/json; charset=UTF-8"}

        t = datetime.datetime.utcnow()
        timestamp = t.strftime("%Y%m%dT%H%M%SZ")

        headers["X-AuroraDNS-Date"] = timestamp
        headers["Authorization"] = self.gen_auth_header(
            self.user_id, self.key, method, action, timestamp
        )

        return super().request(
            action=action, params=params, data=data, method=method, headers=headers
        )


class AuroraDNSDriver(DNSDriver):
    name = "AuroraDNS"
    website = "https://www.pcextreme.nl/en/aurora/dns"
    connectionCls = AuroraDNSConnection

    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.MX: "MX",
        RecordType.NS: "NS",
        RecordType.SOA: "SOA",
        RecordType.SRV: "SRV",
        RecordType.TXT: "TXT",
        RecordType.DS: "DS",
        RecordType.PTR: "PTR",
        RecordType.SSHFP: "SSHFP",
        RecordType.TLSA: "TLSA",
    }

    HEALTHCHECK_TYPE_MAP = {
        AuroraDNSHealthCheckType.HTTP: "HTTP",
        AuroraDNSHealthCheckType.HTTPS: "HTTPS",
        AuroraDNSHealthCheckType.TCP: "TCP",
    }

    def iterate_zones(self):
        res = self.connection.request("/zones")
        for zone in res.parse_body():
            yield self.__res_to_zone(zone)

    def iterate_records(self, zone):
        self.connection.set_context({"resource": "zone", "id": zone.id})
        res = self.connection.request("/zones/%s/records" % zone.id)

        for record in res.parse_body():
            yield self.__res_to_record(zone, record)

    def get_zone(self, zone_id):
        self.connection.set_context({"resource": "zone", "id": zone_id})
        res = self.connection.request("/zones/%s" % zone_id)
        zone = res.parse_body()
        return self.__res_to_zone(zone)

    def get_record(self, zone_id, record_id):
        self.connection.set_context({"resource": "record", "id": record_id})
        res = self.connection.request("/zones/{}/records/{}".format(zone_id, record_id))
        record = res.parse_body()

        zone = self.get_zone(zone_id)

        return self.__res_to_record(zone, record)

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        self.connection.set_context({"resource": "zone", "id": domain})
        res = self.connection.request("/zones", method="POST", data=json.dumps({"name": domain}))
        zone = res.parse_body()
        return self.__res_to_zone(zone)

    def create_record(self, name, zone, type, data, extra=None):
        if name is None:
            name = ""

        rdata = {"name": name, "type": self.RECORD_TYPE_MAP[type], "content": data}

        rdata = self.__merge_extra_data(rdata, extra)

        if "ttl" not in rdata:
            rdata["ttl"] = DEFAULT_ZONE_TTL

        self.connection.set_context({"resource": "zone", "id": zone.id})
        res = self.connection.request(
            "/zones/%s/records" % zone.id, method="POST", data=json.dumps(rdata)
        )

        record = res.parse_body()
        return self.__res_to_record(zone, record)

    def delete_zone(self, zone):
        self.connection.set_context({"resource": "zone", "id": zone.id})
        self.connection.request("/zones/%s" % zone.id, method="DELETE")
        return True

    def delete_record(self, record):
        self.connection.set_context({"resource": "record", "id": record.id})
        self.connection.request(
            "/zones/{}/records/{}".format(record.zone.id, record.id), method="DELETE"
        )
        return True

    def list_record_types(self):
        types = []
        for record_type in self.RECORD_TYPE_MAP.keys():
            types.append(record_type)

        return types

    def update_record(self, record, name, type, data, extra=None):
        rdata = {}

        if name is not None:
            rdata["name"] = name

        if type is not None:
            rdata["type"] = self.RECORD_TYPE_MAP[type]

        if data is not None:
            rdata["content"] = data

        rdata = self.__merge_extra_data(rdata, extra)

        self.connection.set_context({"resource": "record", "id": record.id})
        self.connection.request(
            "/zones/{}/records/{}".format(record.zone.id, record.id),
            method="PUT",
            data=json.dumps(rdata),
        )

        return self.get_record(record.zone.id, record.id)

    def ex_list_healthchecks(self, zone):
        """
        List all Health Checks in a zone.

        :param zone: Zone to list health checks for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`AuroraDNSHealthCheck`
        """
        healthchecks = []
        self.connection.set_context({"resource": "zone", "id": zone.id})
        res = self.connection.request("/zones/%s/health_checks" % zone.id)

        for healthcheck in res.parse_body():
            healthchecks.append(self.__res_to_healthcheck(zone, healthcheck))

        return healthchecks

    def ex_get_healthcheck(self, zone, health_check_id):
        """
        Get a single Health Check from a zone

        :param zone: Zone in which the health check is
        :type zone: :class:`Zone`

        :param health_check_id: ID of the required health check
        :type  health_check_id: ``str``

        :return: :class:`AuroraDNSHealthCheck`
        """
        self.connection.set_context({"resource": "healthcheck", "id": health_check_id})
        res = self.connection.request("/zones/{}/health_checks/{}".format(zone.id, health_check_id))
        check = res.parse_body()

        return self.__res_to_healthcheck(zone, check)

    def ex_create_healthcheck(
        self,
        zone,
        type,
        hostname,
        port,
        path,
        interval,
        threshold,
        ipaddress=None,
        enabled=True,
        extra=None,
    ):
        """
        Create a new Health Check in a zone

        :param zone: Zone in which the health check should be created
        :type zone: :class:`Zone`

        :param type: The type of health check to be created
        :type  type: :class:`AuroraDNSHealthCheckType`

        :param hostname: The hostname of the target to monitor
        :type  hostname: ``str``

        :param port: The port of the target to monitor. E.g. 80 for HTTP
        :type  port: ``int``

        :param path: The path of the target to monitor. Only used by HTTP
                     at this moment. Usually this is simple /.
        :type  path: ``str``

        :param interval: The interval of checks. 10, 30 or 60 seconds.
        :type  interval: ``int``

        :param threshold: The threshold of failures before the healthcheck is
                          marked as failed.
        :type  threshold: ``int``

        :param ipaddress: (optional) The IP Address of the target to monitor.
                          You can pass a empty string if this is not required.
        :type  ipaddress: ``str``

        :param enabled: (optional) If this healthcheck is enabled to run
        :type  enabled: ``bool``

        :param extra: (optional) Extra attributes (driver specific).
        :type  extra: ``dict``

        :return: :class:`AuroraDNSHealthCheck`
        """
        cdata = {
            "type": self.HEALTHCHECK_TYPE_MAP[type],
            "hostname": hostname,
            "ipaddress": ipaddress,
            "port": int(port),
            "interval": int(interval),
            "path": path,
            "threshold": int(threshold),
            "enabled": enabled,
        }

        self.connection.set_context({"resource": "zone", "id": zone.id})
        res = self.connection.request(
            "/zones/%s/health_checks" % zone.id, method="POST", data=json.dumps(cdata)
        )

        healthcheck = res.parse_body()
        return self.__res_to_healthcheck(zone, healthcheck)

    def ex_update_healthcheck(
        self,
        healthcheck,
        type=None,
        hostname=None,
        ipaddress=None,
        port=None,
        path=None,
        interval=None,
        threshold=None,
        enabled=None,
        extra=None,
    ):
        """
        Update an existing Health Check

        :param zone: The healthcheck which has to be updated
        :type zone: :class:`AuroraDNSHealthCheck`

        :param type: (optional) The type of health check to be created
        :type  type: :class:`AuroraDNSHealthCheckType`

        :param hostname: (optional) The hostname of the target to monitor
        :type  hostname: ``str``

        :param ipaddress: (optional) The IP Address of the target to monitor.
                          You can pass a empty string if this is not required.
        :type  ipaddress: ``str``

        :param port: (optional) The port of the target to monitor. E.g. 80
                     for HTTP
        :type  port: ``int``

        :param path: (optional) The path of the target to monitor.
                     Only used by HTTP at this moment. Usually just '/'.
        :type  path: ``str``

        :param interval: (optional) The interval of checks.
                         10, 30 or 60 seconds.
        :type  interval: ``int``

        :param threshold: (optional) The threshold of failures before the
                          healthcheck is marked as failed.
        :type  threshold: ``int``

        :param enabled: (optional) If this healthcheck is enabled to run
        :type  enabled: ``bool``

        :param extra: (optional) Extra attributes (driver specific).
        :type  extra: ``dict``

        :return: :class:`AuroraDNSHealthCheck`
        """
        cdata = {}

        if type is not None:
            cdata["type"] = self.HEALTHCHECK_TYPE_MAP[type]

        if hostname is not None:
            cdata["hostname"] = hostname

        if ipaddress is not None:
            if len(ipaddress) == 0:
                cdata["ipaddress"] = None
            else:
                cdata["ipaddress"] = ipaddress

        if port is not None:
            cdata["port"] = int(port)

        if path is not None:
            cdata["path"] = path

        if interval is not None:
            cdata["interval"] = int(interval)

        if threshold is not None:
            cdata["threshold"] = threshold

        if enabled is not None:
            cdata["enabled"] = bool(enabled)

        self.connection.set_context({"resource": "healthcheck", "id": healthcheck.id})

        self.connection.request(
            "/zones/{}/health_checks/{}".format(healthcheck.zone.id, healthcheck.id),
            method="PUT",
            data=json.dumps(cdata),
        )

        return self.ex_get_healthcheck(healthcheck.zone, healthcheck.id)

    def ex_delete_healthcheck(self, healthcheck):
        """
        Remove an existing Health Check

        :param zone: The healthcheck which has to be removed
        :type zone: :class:`AuroraDNSHealthCheck`
        """
        self.connection.set_context({"resource": "healthcheck", "id": healthcheck.id})

        self.connection.request(
            "/zones/{}/health_checks/{}".format(healthcheck.zone.id, healthcheck.id),
            method="DELETE",
        )
        return True

    def __res_to_record(self, zone, record):
        if len(record["name"]) == 0:
            name = None
        else:
            name = record["name"]

        extra = {}
        extra["created"] = record["created"]
        extra["modified"] = record["modified"]
        extra["disabled"] = record["disabled"]
        extra["ttl"] = record["ttl"]
        extra["priority"] = record["prio"]

        return Record(
            id=record["id"],
            name=name,
            type=record["type"],
            data=record["content"],
            zone=zone,
            driver=self.connection.driver,
            ttl=record["ttl"],
            extra=extra,
        )

    def __res_to_zone(self, zone):
        return Zone(
            id=zone["id"],
            domain=zone["name"],
            type=DEFAULT_ZONE_TYPE,
            ttl=DEFAULT_ZONE_TTL,
            driver=self.connection.driver,
            extra={
                "created": zone["created"],
                "servers": zone["servers"],
                "account_id": zone["account_id"],
                "cluster_id": zone["cluster_id"],
            },
        )

    def __res_to_healthcheck(self, zone, healthcheck):
        return AuroraDNSHealthCheck(
            id=healthcheck["id"],
            type=healthcheck["type"],
            hostname=healthcheck["hostname"],
            ipaddress=healthcheck["ipaddress"],
            health=healthcheck["health"],
            threshold=healthcheck["threshold"],
            path=healthcheck["path"],
            interval=healthcheck["interval"],
            port=healthcheck["port"],
            enabled=healthcheck["enabled"],
            zone=zone,
            driver=self.connection.driver,
        )

    def __merge_extra_data(self, rdata, extra):
        if extra is not None:
            for param in VALID_RECORD_PARAMS_EXTRA:
                if param in extra:
                    rdata[param] = extra[param]

        return rdata



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["SLB_API_VERSION", "SLBDriver"]

try:
    import simplejson as json
except ImportError:
    import json

from libcloud.utils.py3 import u
from libcloud.utils.xml import findall, findattr, findtext
from libcloud.utils.misc import ReprMixin
from libcloud.common.types import LibcloudError
from libcloud.common.aliyun import AliyunXmlResponse, SignedAliyunConnection
from libcloud.loadbalancer.base import Driver, Member, Algorithm, LoadBalancer
from libcloud.loadbalancer.types import State

SLB_API_VERSION = "2014-05-15"
SLB_API_HOST = "slb.aliyuncs.com"
DEFAULT_SIGNATURE_VERSION = "1.0"


STATE_MAPPINGS = {
    "inactive": State.UNKNOWN,
    "active": State.RUNNING,
    "locked": State.PENDING,
}


RESOURCE_EXTRA_ATTRIBUTES_MAP = {
    "balancer": {
        "create_timestamp": {"xpath": "CreateTimeStamp", "transform_func": int},
        "address_type": {"xpath": "AddressType", "transform_func": u},
        "region_id": {"xpath": "RegionId", "transform_func": u},
        "region_id_alias": {"xpath": "RegionIdAlias", "transform_func": u},
        "create_time": {"xpath": "CreateTime", "transform_func": u},
        "master_zone_id": {"xpath": "MasterZoneId", "transform_func": u},
        "slave_zone_id": {"xpath": "SlaveZoneId", "transform_func": u},
        "network_type": {"xpath": "NetworkType", "transform_func": u},
    }
}


SLB_SCHEDULER_TO_ALGORITHM = {
    "wrr": Algorithm.WEIGHTED_ROUND_ROBIN,
    "wlc": Algorithm.WEIGHTED_LEAST_CONNECTIONS,
}


ALGORITHM_TO_SLB_SCHEDULER = {
    Algorithm.WEIGHTED_ROUND_ROBIN: "wrr",
    Algorithm.WEIGHTED_LEAST_CONNECTIONS: "wlc",
}


class SLBConnection(SignedAliyunConnection):
    api_version = SLB_API_VERSION
    host = SLB_API_HOST
    responseCls = AliyunXmlResponse
    service_name = "slb"


class SLBLoadBalancerAttribute:
    """
    This class used to get listeners and backend servers related to a balancer
    listeners is a ``list`` of ``dict``, each element contains
    'ListenerPort' and 'ListenerProtocol' keys.
    backend_servers is a ``list`` of ``dict``, each element contains
    'ServerId' and 'Weight' keys.
    """

    def __init__(self, balancer, listeners, backend_servers, extra=None):
        self.balancer = balancer
        self.listeners = listeners or []
        self.backend_servers = backend_servers or []
        self.extra = extra or {}

    def is_listening(self, port):
        for listener in self.listeners:
            if listener.get("ListenerPort") == port:
                return True
        return False

    def is_attached(self, member):
        for server in self.backend_servers:
            if server.get("Serverid") == member.id:
                return True
        return False

    def __repr__(self):
        return "<SLBLoadBalancerAttribute id={}, ports={}, servers={} ...>".format(
            self.balancer.id,
            self.listeners,
            self.backend_servers,
        )


class SLBLoadBalancerListener(ReprMixin):
    """
    Base SLB load balancer listener class
    """

    _repr_attributes = ["port", "backend_port", "scheduler", "bandwidth"]
    action = None
    option_keys = []

    def __init__(self, port, backend_port, algorithm, bandwidth, extra=None):
        self.port = port
        self.backend_port = backend_port
        self.scheduler = ALGORITHM_TO_SLB_SCHEDULER.get(algorithm, "wrr")
        self.bandwidth = bandwidth
        self.extra = extra or {}

    @classmethod
    def create(cls, port, backend_port, algorithm, bandwidth, extra=None):
        return cls(port, backend_port, algorithm, bandwidth, extra=extra)

    def get_create_params(self):
        params = self.get_required_params()
        options = self.get_optional_params()
        options.update(params)
        return options

    def get_required_params(self):
        params = {
            "Action": self.action,
            "ListenerPort": self.port,
            "BackendServerPort": self.backend_port,
            "Scheduler": self.scheduler,
            "Bandwidth": self.bandwidth,
        }
        return params

    def get_optional_params(self):
        options = {}
        for option in self.option_keys:
            if self.extra and option in self.extra:
                options[option] = self.extra[option]
        return options


class SLBLoadBalancerHttpListener(SLBLoadBalancerListener):
    """
    This class represents a rule to route http request to the backends.
    """

    action = "CreateLoadBalancerHTTPListener"
    option_keys = [
        "XForwardedFor",
        "StickySessionType",
        "CookieTimeout",
        "Cookie",
        "HealthCheckDomain",
        "HealthCheckURI",
        "HealthCheckConnectPort",
        "HealthyThreshold",
        "UnhealthyThreshold",
        "HealthCheckTimeout",
        "HealthCheckInterval",
        "HealthCheckHttpCode",
    ]

    def __init__(
        self,
        port,
        backend_port,
        algorithm,
        bandwidth,
        sticky_session,
        health_check,
        extra=None,
    ):
        super().__init__(port, backend_port, algorithm, bandwidth, extra=extra)
        self.sticky_session = sticky_session
        self.health_check = health_check

    def get_required_params(self):
        params = super().get_required_params()
        params["StickySession"] = self.sticky_session
        params["HealthCheck"] = self.health_check
        return params

    @classmethod
    def create(cls, port, backend_port, algorithm, bandwidth, extra={}):
        if "StickySession" not in extra:
            raise AttributeError("StickySession is required")
        if "HealthCheck" not in extra:
            raise AttributeError("HealthCheck is required")
        sticky_session = extra["StickySession"]
        health_check = extra["HealthCheck"]
        return cls(
            port,
            backend_port,
            algorithm,
            bandwidth,
            sticky_session,
            health_check,
            extra=extra,
        )


class SLBLoadBalancerHttpsListener(SLBLoadBalancerListener):
    """
    This class represents a rule to route https request to the backends.
    """

    action = "CreateLoadBalancerHTTPSListener"
    option_keys = [
        "XForwardedFor",
        "StickySessionType",
        "CookieTimeout",
        "Cookie",
        "HealthCheckDomain",
        "HealthCheckURI",
        "HealthCheckConnectPort",
        "HealthyThreshold",
        "UnhealthyThreshold",
        "HealthCheckTimeout",
        "HealthCheckInterval",
        "HealthCheckHttpCode",
    ]

    def __init__(
        self,
        port,
        backend_port,
        algorithm,
        bandwidth,
        sticky_session,
        health_check,
        certificate_id,
        extra=None,
    ):
        super().__init__(port, backend_port, algorithm, bandwidth, extra=extra)
        self.sticky_session = sticky_session
        self.health_check = health_check
        self.certificate_id = certificate_id

    def get_required_params(self):
        params = super().get_required_params()
        params["StickySession"] = self.sticky_session
        params["HealthCheck"] = self.health_check
        params["ServerCertificateId"] = self.certificate_id
        return params

    @classmethod
    def create(cls, port, backend_port, algorithm, bandwidth, extra={}):
        if "StickySession" not in extra:
            raise AttributeError("StickySession is required")
        if "HealthCheck" not in extra:
            raise AttributeError("HealthCheck is required")
        if "ServerCertificateId" not in extra:
            raise AttributeError("ServerCertificateId is required")
        sticky_session = extra["StickySession"]
        health_check = extra["HealthCheck"]
        certificate_id = extra["ServerCertificateId"]
        return cls(
            port,
            backend_port,
            algorithm,
            bandwidth,
            sticky_session,
            health_check,
            certificate_id,
            extra=extra,
        )


class SLBLoadBalancerTcpListener(SLBLoadBalancerListener):
    """
    This class represents a rule to route tcp request to the backends.
    """

    action = "CreateLoadBalancerTCPListener"
    option_keys = [
        "PersistenceTimeout",
        "HealthCheckType",
        "HealthCheckDomain",
        "HealthCheckURI",
        "HealthCheckConnectPort",
        "HealthyThreshold",
        "UnhealthyThreshold",
        "HealthCheckConnectTimeout",
        "HealthCheckInterval",
        "HealthCheckHttpCode",
    ]


class SLBLoadBalancerUdpListener(SLBLoadBalancerTcpListener):
    """
    This class represents a rule to route udp request to the backends.
    """

    action = "CreateLoadBalancerUDPListener"
    option_keys = [
        "PersistenceTimeout",
        "HealthCheckConnectPort",
        "HealthyThreshold",
        "UnhealthyThreshold",
        "HealthCheckConnectTimeout",
        "HealthCheckInterval",
    ]


class SLBServerCertificate(ReprMixin):
    _repr_attributes = ["id", "name", "fingerprint"]

    def __init__(self, id, name, fingerprint):
        self.id = id
        self.name = name
        self.fingerprint = fingerprint


PROTOCOL_TO_LISTENER_MAP = {
    "http": SLBLoadBalancerHttpListener,
    "https": SLBLoadBalancerHttpsListener,
    "tcp": SLBLoadBalancerTcpListener,
    "udp": SLBLoadBalancerUdpListener,
}


class SLBDriver(Driver):
    """
    Aliyun SLB load balancer driver.
    """

    name = "Aliyun Server Load Balancer"
    website = "https://www.aliyun.com/product/slb"
    connectionCls = SLBConnection
    path = "/"
    namespace = None

    _VALUE_TO_ALGORITHM_MAP = SLB_SCHEDULER_TO_ALGORITHM

    _ALGORITHM_TO_VALUE_MAP = ALGORITHM_TO_SLB_SCHEDULER

    def __init__(self, access_id, secret, region):
        super().__init__(access_id, secret)
        self.region = region

    def list_protocols(self):
        return list(PROTOCOL_TO_LISTENER_MAP.keys())

    def list_balancers(self, ex_balancer_ids=None, ex_filters=None):
        """
        List all loadbalancers

        @inherits :class:`Driver.list_balancers`

        :keyword ex_balancer_ids: a list of balancer ids to filter results
                                  Only balancers which's id in this list
                                  will be returned
        :type ex_balancer_ids: ``list`` of ``str``

        :keyword ex_filters: attributes to filter results. Only balancers
                             which have all the desired attributes
                             and values will be returned
        :type ex_filters: ``dict``
        """

        params = {"Action": "DescribeLoadBalancers", "RegionId": self.region}
        if ex_balancer_ids and isinstance(ex_balancer_ids, list):
            params["LoadBalancerId"] = ",".join(ex_balancer_ids)

        if ex_filters and isinstance(ex_filters, dict):
            ex_filters.update(params)
            params = ex_filters
        resp_body = self.connection.request(self.path, params=params).object
        return self._to_balancers(resp_body)

    def create_balancer(
        self,
        name,
        port,
        protocol,
        algorithm,
        members,
        ex_bandwidth=None,
        ex_internet_charge_type=None,
        ex_address_type=None,
        ex_vswitch_id=None,
        ex_master_zone_id=None,
        ex_slave_zone_id=None,
        ex_client_token=None,
        **kwargs,
    ):
        """
        Create a new load balancer instance

        @inherits: :class:`Driver.create_balancer`

        :keyword ex_bandwidth: The max bandwidth limit for `paybybandwidth`
                               internet charge type, in Mbps unit
        :type ex_bandwidth: ``int`` in range [1, 1000]

        :keyword ex_internet_charge_type: The internet charge type
        :type ex_internet_charge_type: a ``str`` of `paybybandwidth`
                                       or `paybytraffic`

        :keyword ex_address_type: The listening IP address type
        :type ex_address_type: a ``str`` of `internet` or `intranet`

        :keyword ex_vswitch_id: The vswitch id in a VPC network
        :type ex_vswitch_id: ``str``

        :keyword ex_master_zone_id: The id of the master availability zone
        :type ex_master_zone_id: ``str``

        :keyword ex_slave_zone_id: The id of the slave availability zone
        :type ex_slave_zone_id: ``str``

        :keyword ex_client_token: The token generated by client to
                                  identify requests
        :type ex_client_token: ``str``
        """

        # 1.Create load balancer
        params = {"Action": "CreateLoadBalancer", "RegionId": self.region}
        if name:
            params["LoadBalancerName"] = name
        if not port:
            raise AttributeError("port is required")
        if not protocol:
            # NOTE(samsong8610): Use http listener as default
            protocol = "http"
        if protocol not in PROTOCOL_TO_LISTENER_MAP:
            raise AttributeError("unsupported protocol %s" % protocol)

        # Bandwidth in range [1, 1000] Mbps
        bandwidth = -1
        if ex_bandwidth:
            try:
                bandwidth = int(ex_bandwidth)
            except ValueError:
                raise AttributeError("ex_bandwidth should be a integer in " "range [1, 1000].")
            params["Bandwidth"] = bandwidth

        if ex_internet_charge_type:
            if ex_internet_charge_type.lower() == "paybybandwidth":
                if bandwidth == -1:
                    raise AttributeError(
                        "PayByBandwidth internet charge type" " need ex_bandwidth be set"
                    )
            params["InternetChargeType"] = ex_internet_charge_type

        if ex_address_type:
            if ex_address_type.lower() not in ("internet", "intranet"):
                raise AttributeError('ex_address_type should be "internet" ' 'or "intranet"')
            params["AddressType"] = ex_address_type

        if ex_vswitch_id:
            params["VSwitchId"] = ex_vswitch_id

        if ex_master_zone_id:
            params["MasterZoneId"] = ex_master_zone_id
        if ex_slave_zone_id:
            params["SlaveZoneId"] = ex_slave_zone_id

        if ex_client_token:
            params["ClientToken"] = ex_client_token

        if members and isinstance(members, list):
            backend_ports = [member.port for member in members]
            if len(set(backend_ports)) != 1:
                raise AttributeError("the ports of members should be unique")
            # NOTE(samsong8610): If members do not provide backend port,
            #                    default to listening port
            backend_port = backend_ports[0] or port
        else:
            backend_port = port

        balancer = None
        try:
            resp_body = self.connection.request(self.path, params).object
            balancer = self._to_balancer(resp_body)
            balancer.port = port

            # 2.Add backend servers
            if members is None:
                members = []
            for member in members:
                self.balancer_attach_member(balancer, member)
            # 3.Create listener
            # NOTE(samsong8610): Assume only create a listener which uses all
            #                    the bandwidth.
            self.ex_create_listener(
                balancer, backend_port, protocol, algorithm, bandwidth, **kwargs
            )
            self.ex_start_listener(balancer, port)
            return balancer
        except Exception as e:
            if balancer is not None:
                try:
                    self.destroy_balancer(balancer)
                except Exception:
                    pass
            raise e

    def destroy_balancer(self, balancer):
        params = {"Action": "DeleteLoadBalancer", "LoadBalancerId": balancer.id}
        resp = self.connection.request(self.path, params)
        return resp.success()

    def get_balancer(self, balancer_id):
        balancers = self.list_balancers(ex_balancer_ids=[balancer_id])
        if len(balancers) != 1:
            raise LibcloudError("could not find load balancer with id %s" % balancer_id)
        return balancers[0]

    def balancer_attach_compute_node(self, balancer, node):
        if len(node.public_ips) > 0:
            ip = node.public_ips[0]
        else:
            ip = node.private_ips[0]
        member = Member(id=node.id, ip=ip, port=balancer.port)
        return self.balancer_attach_member(balancer, member)

    def balancer_attach_member(self, balancer, member):
        params = {"Action": "AddBackendServers", "LoadBalancerId": balancer.id}
        if member and isinstance(member, Member):
            params["BackendServers"] = self._to_servers_json([member])
        self.connection.request(self.path, params)
        return member

    def balancer_detach_member(self, balancer, member):
        params = {"Action": "RemoveBackendServers", "LoadBalancerId": balancer.id}
        if member and isinstance(member, Member):
            params["BackendServers"] = self._list_to_json([member.id])
        self.connection.request(self.path, params)
        return member

    def balancer_list_members(self, balancer):
        attribute = self.ex_get_balancer_attribute(balancer)
        members = [
            Member(
                server["ServerId"],
                None,
                None,
                balancer=balancer,
                extra={"Weight": server["Weight"]},
            )
            for server in attribute.backend_servers
        ]
        return members

    def ex_get_balancer_attribute(self, balancer):
        """
        Get balancer attribute

        :param balancer: the balancer to get attribute
        :type balancer: ``LoadBalancer``

        :return: the balancer attribute
        :rtype: ``SLBLoadBalancerAttribute``
        """

        params = {
            "Action": "DescribeLoadBalancerAttribute",
            "LoadBalancerId": balancer.id,
        }
        resp_body = self.connection.request(self.path, params).object
        attribute = self._to_balancer_attribute(resp_body)
        return attribute

    def ex_list_listeners(self, balancer):
        """
        Get all listener related to the given balancer

        :param balancer: the balancer to list listeners
        :type balancer: ``LoadBalancer``

        :return: a list of listeners
        :rtype: ``list`` of ``SLBLoadBalancerListener``
        """

        attribute = self.ex_get_balancer_attribute(balancer)
        listeners = [
            SLBLoadBalancerListener(each["ListenerPort"], None, None, None)
            for each in attribute.listeners
        ]
        return listeners

    def ex_create_listener(self, balancer, backend_port, protocol, algorithm, bandwidth, **kwargs):
        """
        Create load balancer listening rule.

        :param balancer: the balancer which the rule belongs to.
                         The listener created will listen on the port of the
                         the balancer as default. 'ListenerPort' in kwargs
                         will *OVERRIDE* it.
        :type balancer: ``LoadBalancer``

        :param backend_port: the backend server port
        :type backend_port: ``int``

        :param protocol: the balancer protocol, default to http
        :type protocol: ``str``

        :param algorithm: the balancer routing algorithm
        :type algorithm: ``Algorithm``

        :param bandwidth: the listener bandwidth limits
        :type bandwidth: ``str``

        :return: the created listener
        :rtype: ``SLBLoadBalancerListener``
        """

        cls = PROTOCOL_TO_LISTENER_MAP.get(protocol, SLBLoadBalancerHttpListener)
        if "ListenerPort" in kwargs:
            port = kwargs["ListenerPort"]
        else:
            port = balancer.port
        listener = cls.create(port, backend_port, algorithm, bandwidth, extra=kwargs)
        params = listener.get_create_params()
        params["LoadBalancerId"] = balancer.id
        params["RegionId"] = self.region
        resp = self.connection.request(self.path, params)
        return resp.success()

    def ex_start_listener(self, balancer, port):
        """
        Start balancer's listener listening the given port.

        :param balancer: a load balancer
        :type balancer: ``LoadBalancer``

        :param port: listening port
        :type port: ``int``

        :return: whether operation is success
        :rtype: ``bool``
        """

        params = {
            "Action": "StartLoadBalancerListener",
            "LoadBalancerId": balancer.id,
            "ListenerPort": port,
        }
        resp = self.connection.request(self.path, params)
        return resp.success()

    def ex_stop_listener(self, balancer, port):
        """
        Stop balancer's listener listening the given port.

        :param balancer: a load balancer
        :type balancer: ``LoadBalancer``

        :param port: listening port
        :type port: ``int``

        :return: whether operation is success
        :rtype: ``bool``
        """

        params = {
            "Action": "StopLoadBalancerListener",
            "LoadBalancerId": balancer.id,
            "ListenerPort": port,
        }
        resp = self.connection.request(self.path, params)
        return resp.success()

    def ex_upload_certificate(self, name, server_certificate, private_key):
        """
        Upload certificate and private key for https load balancer listener

        :param name: the certificate name
        :type name: ``str``

        :param server_certificate: the content of the certificate to upload
                                   in PEM format
        :type server_certificate: ``str``

        :param private_key: the content of the private key to upload
                            in PEM format
        :type private_key: ``str``

        :return: new created certificate info
        :rtype: ``SLBServerCertificate``
        """

        params = {
            "Action": "UploadServerCertificate",
            "RegionId": self.region,
            "ServerCertificate": server_certificate,
            "PrivateKey": private_key,
        }
        if name:
            params["ServerCertificateName"] = name
        resp_body = self.connection.request(self.path, params).object
        return self._to_server_certificate(resp_body)

    def ex_list_certificates(self, certificate_ids=[]):
        """
        List all server certificates

        :param certificate_ids: certificate ids to filter results
        :type certificate_ids: ``str``

        :return: certificates
        :rtype: ``SLBServerCertificate``
        """

        params = {"Action": "DescribeServerCertificates", "RegionId": self.region}
        if certificate_ids and isinstance(certificate_ids, list):
            params["ServerCertificateId"] = ",".join(certificate_ids)

        resp_body = self.connection.request(self.path, params).object
        cert_elements = findall(
            resp_body, "ServerCertificates/ServerCertificate", namespace=self.namespace
        )
        certificates = [self._to_server_certificate(el) for el in cert_elements]
        return certificates

    def ex_delete_certificate(self, certificate_id):
        """
        Delete the given server certificate

        :param certificate_id: the id of the certificate to delete
        :type certificate_id: ``str``

        :return: whether process is success
        :rtype: ``bool``
        """

        params = {
            "Action": "DeleteServerCertificate",
            "RegionId": self.region,
            "ServerCertificateId": certificate_id,
        }
        resp = self.connection.request(self.path, params)
        return resp.success()

    def ex_set_certificate_name(self, certificate_id, name):
        """
        Set server certificate name.

        :param certificate_id: the id of the server certificate to update
        :type certificate_id: ``str``

        :param name: the new name
        :type name: ``str``

        :return: whether updating is success
        :rtype: ``bool``
        """

        params = {
            "Action": "SetServerCertificateName",
            "RegionId": self.region,
            "ServerCertificateId": certificate_id,
            "ServerCertificateName": name,
        }
        resp = self.connection.request(self.path, params)
        return resp.success()

    def _to_balancers(self, element):
        xpath = "LoadBalancers/LoadBalancer"
        return [
            self._to_balancer(el)
            for el in findall(element=element, xpath=xpath, namespace=self.namespace)
        ]

    def _to_balancer(self, el):
        _id = findtext(element=el, xpath="LoadBalancerId", namespace=self.namespace)
        name = findtext(element=el, xpath="LoadBalancerName", namespace=self.namespace)
        status = findtext(element=el, xpath="LoadBalancerStatus", namespace=self.namespace)
        state = STATE_MAPPINGS.get(status, State.UNKNOWN)
        address = findtext(element=el, xpath="Address", namespace=self.namespace)
        extra = self._get_extra_dict(el, RESOURCE_EXTRA_ATTRIBUTES_MAP["balancer"])

        balancer = LoadBalancer(
            id=_id,
            name=name,
            state=state,
            ip=address,
            port=None,
            driver=self,
            extra=extra,
        )
        return balancer

    def _create_list_params(self, params, items, label):
        """
        return parameter list
        """
        if isinstance(items, str):
            items = [items]
        for index, item in enumerate(items):
            params[label % (index + 1)] = item
        return params

    def _get_extra_dict(self, element, mapping):
        """
        Extract attributes from the element based on rules provided in the
        mapping dictionary.

        :param      element: Element to parse the values from.
        :type       element: xml.etree.ElementTree.Element.

        :param      mapping: Dictionary with the extra layout
        :type       node: :class:`Node`

        :rtype: ``dict``
        """
        extra = {}
        for attribute, values in mapping.items():
            transform_func = values["transform_func"]
            value = findattr(element=element, xpath=values["xpath"], namespace=self.namespace)
            if value:
                try:
                    extra[attribute] = transform_func(value)
                except Exception:
                    extra[attribute] = None
            else:
                extra[attribute] = value

        return extra

    def _to_servers_json(self, members):
        servers = []
        for each in members:
            server = {"ServerId": each.id, "Weight": "100"}
            if "Weight" in each.extra:
                server["Weight"] = each.extra["Weight"]
            servers.append(server)
        try:
            return json.dumps(servers)
        except Exception:
            raise AttributeError("could not convert member to backend server")

    def _to_balancer_attribute(self, element):
        balancer = self._to_balancer(element)
        port_proto_elements = findall(
            element,
            "ListenerPortsAndProtocol/ListenerPortAndProtocol",
            namespace=self.namespace,
        )
        if len(port_proto_elements) > 0:
            listeners = [self._to_port_and_protocol(el) for el in port_proto_elements]
        else:
            port_elements = findall(element, "ListenerPorts/ListenerPort", namespace=self.namespace)
            listeners = [
                {"ListenerPort": el.text, "ListenerProtocol": "http"} for el in port_elements
            ]
        server_elements = findall(element, "BackendServers/BackendServer", namespace=self.namespace)
        backend_servers = [self._to_server_and_weight(el) for el in server_elements]
        return SLBLoadBalancerAttribute(balancer, listeners, backend_servers)

    def _to_port_and_protocol(self, el):
        port = findtext(el, "ListenerPort", namespace=self.namespace)
        protocol = findtext(el, "ListenerProtocol", namespace=self.namespace)
        return {"ListenerPort": port, "ListenerProtocol": protocol}

    def _to_server_and_weight(self, el):
        server_id = findtext(el, "ServerId", namespace=self.namespace)
        weight = findtext(el, "Weight", namespace=self.namespace)
        return {"ServerId": server_id, "Weight": weight}

    def _to_server_certificate(self, el):
        _id = findtext(el, "ServerCertificateId", namespace=self.namespace)
        name = findtext(el, "ServerCertificateName", namespace=self.namespace)
        fingerprint = findtext(el, "Fingerprint", namespace=self.namespace)
        return SLBServerCertificate(id=_id, name=name, fingerprint=fingerprint)

    def _list_to_json(self, value):
        try:
            return json.dumps(value)
        except Exception:
            return "[]"



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
DigitalOcean Driver
"""

import json
import warnings

from libcloud.utils.py3 import httplib
from libcloud.common.types import InvalidCredsError
from libcloud.compute.base import (
    Node,
    KeyPair,
    NodeSize,
    NodeImage,
    NodeDriver,
    NodeLocation,
    StorageVolume,
    VolumeSnapshot,
)
from libcloud.compute.types import Provider, NodeState
from libcloud.utils.iso8601 import parse_date
from libcloud.common.digitalocean import DigitalOcean_v1_Error, DigitalOcean_v2_BaseDriver

__all__ = ["DigitalOceanNodeDriver", "DigitalOcean_v2_NodeDriver"]


class DigitalOceanNodeDriver(NodeDriver):
    """
    DigitalOcean NodeDriver defaulting to using APIv2.

    :keyword    key: Personal Access Token required for authentication.
    :type       key: ``str``

    :keyword    secret: Previously used with API version ``v1``. (deprecated)
    :type       secret: ``str``

    :keyword    api_version: Specifies the API version to use. Defaults to
                             using ``v2``, currently the only valid option.
                             (optional)
    :type       api_version: ``str``
    """

    type = Provider.DIGITAL_OCEAN
    name = "DigitalOcean"
    website = "https://www.digitalocean.com"

    def __new__(cls, key, secret=None, api_version="v2", **kwargs):
        if cls is DigitalOceanNodeDriver:
            if api_version == "v1" or secret is not None:
                if secret is not None and api_version == "v2":
                    raise InvalidCredsError("secret not accepted for v2 authentication")
                raise DigitalOcean_v1_Error()
            elif api_version == "v2":
                cls = DigitalOcean_v2_NodeDriver
            else:
                raise NotImplementedError("Unsupported API version: %s" % (api_version))
        return super().__new__(cls, **kwargs)


# TODO Implement v1 driver using KeyPair
class SSHKey:
    def __init__(self, id, name, pub_key):
        self.id = id
        self.name = name
        self.pub_key = pub_key

    def __repr__(self):
        return ("<SSHKey: id=%s, name=%s, pub_key=%s>") % (
            self.id,
            self.name,
            self.pub_key,
        )


class DigitalOcean_v2_NodeDriver(DigitalOcean_v2_BaseDriver, DigitalOceanNodeDriver):
    """
    DigitalOcean NodeDriver using v2 of the API.
    """

    NODE_STATE_MAP = {
        "new": NodeState.PENDING,
        "off": NodeState.STOPPED,
        "active": NodeState.RUNNING,
        "archive": NodeState.TERMINATED,
    }

    EX_CREATE_ATTRIBUTES = ["backups", "ipv6", "private_networking", "tags", "ssh_keys"]

    def list_images(self):
        data = self._paginated_request("/v2/images", "images")
        return list(map(self._to_image, data))

    def list_key_pairs(self):
        """
        List all the available SSH keys.

        :return: Available SSH keys.
        :rtype: ``list`` of :class:`KeyPair`
        """
        data = self._paginated_request("/v2/account/keys", "ssh_keys")
        return list(map(self._to_key_pair, data))

    def list_locations(self, ex_available=True):
        """
        List locations

        :param ex_available: Only return locations which are available.
        :type ex_evailable: ``bool``
        """
        locations = []
        data = self._paginated_request("/v2/regions", "regions")
        for location in data:
            if ex_available:
                if location.get("available"):
                    locations.append(self._to_location(location))
            else:
                locations.append(self._to_location(location))
        return locations

    def list_nodes(self):
        data = self._paginated_request("/v2/droplets", "droplets")
        return list(map(self._to_node, data))

    def list_sizes(self, location=None):
        data = self._paginated_request("/v2/sizes", "sizes")
        sizes = list(map(self._to_size, data))
        if location:
            sizes = [size for size in sizes if location.id in size.extra.get("regions", [])]
        return sizes

    def list_volumes(self):
        data = self._paginated_request("/v2/volumes", "volumes")
        return list(map(self._to_volume, data))

    def create_node(
        self,
        name,
        size,
        image,
        location,
        ex_create_attr=None,
        ex_ssh_key_ids=None,
        ex_user_data=None,
    ):
        """
        Create a node.

        The `ex_create_attr` parameter can include the following dictionary
        key and value pairs:

        * `backups`: ``bool`` defaults to False
        * `ipv6`: ``bool`` defaults to False
        * `private_networking`: ``bool`` defaults to False
        * `tags`: ``list`` of ``str`` tags
        * `user_data`: ``str`` for cloud-config data
        * `ssh_keys`: ``list`` of ``int`` key ids or ``str`` fingerprints

        `ex_create_attr['ssh_keys']` will override `ex_ssh_key_ids` assignment.

        :keyword ex_create_attr: A dictionary of optional attributes for
                                 droplet creation
        :type ex_create_attr: ``dict``

        :keyword ex_ssh_key_ids: A list of ssh key ids which will be added
                                 to the server. (optional)
        :type ex_ssh_key_ids: ``list`` of ``int`` key ids or ``str``
                              key fingerprints

        :keyword    ex_user_data:  User data to be added to the node on create.
                                     (optional)
        :type       ex_user_data:  ``str``

        :return: The newly created node.
        :rtype: :class:`Node`
        """
        attr = {
            "name": name,
            "size": size.name,
            "image": image.id,
            "region": location.id,
            "user_data": ex_user_data,
        }

        if ex_ssh_key_ids:
            warnings.warn(
                "The ex_ssh_key_ids parameter has been deprecated in"
                " favor of the ex_create_attr parameter."
            )
            attr["ssh_keys"] = ex_ssh_key_ids

        ex_create_attr = ex_create_attr or {}
        for key in ex_create_attr.keys():
            if key in self.EX_CREATE_ATTRIBUTES:
                attr[key] = ex_create_attr[key]

        res = self.connection.request("/v2/droplets", data=json.dumps(attr), method="POST")

        data = res.object["droplet"]
        # TODO: Handle this in the response class
        status = res.object.get("status", "OK")
        if status == "ERROR":
            message = res.object.get("message", None)
            error_message = res.object.get("error_message", message)
            raise ValueError("Failed to create node: %s" % (error_message))

        return self._to_node(data=data)

    def destroy_node(self, node):
        res = self.connection.request("/v2/droplets/%s" % (node.id), method="DELETE")
        return res.status == httplib.NO_CONTENT

    def reboot_node(self, node):
        attr = {"type": "reboot"}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def create_image(self, node, name):
        """
        Create an image from a Node.

        @inherits: :class:`NodeDriver.create_image`

        :param node: Node to use as base for image
        :type node: :class:`Node`

        :param node: Name for image
        :type node: ``str``

        :rtype: ``bool``
        """
        attr = {"type": "snapshot", "name": name}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def delete_image(self, image):
        """Delete an image for node.

        @inherits: :class:`NodeDriver.delete_image`

        :param      image: the image to be deleted
        :type       image: :class:`NodeImage`

        :rtype: ``bool``
        """
        res = self.connection.request("/v2/images/%s" % (image.id), method="DELETE")
        return res.status == httplib.NO_CONTENT

    def get_image(self, image_id):
        """
        Get an image based on an image_id

        @inherits: :class:`NodeDriver.get_image`

        :param image_id: Image identifier
        :type image_id: ``int``

        :return: A NodeImage object
        :rtype: :class:`NodeImage`
        """
        data = self._paginated_request("/v2/images/%s" % (image_id), "image")
        return self._to_image(data)

    def ex_change_kernel(self, node, kernel_id):
        attr = {"type": "change_kernel", "kernel": kernel_id}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def ex_enable_ipv6(self, node):
        attr = {"type": "enable_ipv6"}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def ex_rename_node(self, node, name):
        attr = {"type": "rename", "name": name}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def ex_shutdown_node(self, node):
        attr = {"type": "shutdown"}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def ex_hard_reboot(self, node):
        attr = {"type": "power_cycle"}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def ex_power_on_node(self, node):
        attr = {"type": "power_on"}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def ex_rebuild_node(self, node):
        """
        Destroy and rebuild the node using its base image.

        :param node: Node to rebuild
        :type node: :class:`Node`

        :return True if the operation began successfully
        :rtype ``bool``
        """
        attr = {"type": "rebuild", "image": node.extra["image"]["id"]}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def ex_resize_node(self, node, size):
        """
        Resize the node to a different machine size.  Note that some resize
        operations are reversible, and others are irreversible.

        :param node: Node to rebuild
        :type node: :class:`Node`

        :param size: New size for this machine
        :type node: :class:`NodeSize`

        :return True if the operation began successfully
        :rtype ``bool``
        """
        attr = {"type": "resize", "size": size.name}
        res = self.connection.request(
            "/v2/droplets/%s/actions" % (node.id), data=json.dumps(attr), method="POST"
        )
        return res.status == httplib.CREATED

    def create_key_pair(self, name, public_key=""):
        """
        Create a new SSH key.

        :param      name: Key name (required)
        :type       name: ``str``

        :param      public_key: Valid public key string (required)
        :type       public_key: ``str``
        """
        attr = {"name": name, "public_key": public_key}
        res = self.connection.request("/v2/account/keys", method="POST", data=json.dumps(attr))

        data = res.object["ssh_key"]

        return self._to_key_pair(data=data)

    def delete_key_pair(self, key):
        """
        Delete an existing SSH key.

        :param      key: SSH key (required)
        :type       key: :class:`KeyPair`
        """
        key_id = key.extra["id"]
        res = self.connection.request("/v2/account/keys/%s" % (key_id), method="DELETE")
        return res.status == httplib.NO_CONTENT

    def get_key_pair(self, name):
        """
        Retrieve a single key pair.

        :param name: Name of the key pair to retrieve.
        :type name: ``str``

        :rtype: :class:`.KeyPair`
        """
        qkey = [k for k in self.list_key_pairs() if k.name == name][0]
        data = self.connection.request("/v2/account/keys/%s" % qkey.extra["id"]).object["ssh_key"]
        return self._to_key_pair(data=data)

    def create_volume(self, size, name, location=None, snapshot=None):
        """
        Create a new volume.

        :param size: Size of volume in gigabytes (required)
        :type size: ``int``

        :param name: Name of the volume to be created
        :type name: ``str``

        :param location: Which data center to create a volume in. If
                               empty, undefined behavior will be selected.
                               (optional)
        :type location: :class:`.NodeLocation`

        :param snapshot:  Snapshot from which to create the new
                          volume.  (optional)
        :type snapshot: :class:`.VolumeSnapshot`

        :return: The newly created volume.
        :rtype: :class:`StorageVolume`
        """
        attr = {"name": name, "size_gigabytes": size, "region": location.id}

        res = self.connection.request("/v2/volumes", data=json.dumps(attr), method="POST")
        data = res.object["volume"]
        status = res.object.get("status", "OK")
        if status == "ERROR":
            message = res.object.get("message", None)
            error_message = res.object.get("error_message", message)
            raise ValueError("Failed to create volume: %s" % (error_message))

        return self._to_volume(data=data)

    def destroy_volume(self, volume):
        """
        Destroys a storage volume.

        :param volume: Volume to be destroyed
        :type volume: :class:`StorageVolume`

        :rtype: ``bool``
        """
        res = self.connection.request("/v2/volumes/%s" % volume.id, method="DELETE")
        return res.status == httplib.NO_CONTENT

    def attach_volume(self, node, volume, device=None):
        """
        Attaches volume to node.

        :param node: Node to attach volume to.
        :type node: :class:`.Node`

        :param volume: Volume to attach.
        :type volume: :class:`.StorageVolume`

        :param device: Where the device is exposed, e.g. '/dev/sdb'
        :type device: ``str``

        :rytpe: ``bool``
        """
        attr = {
            "type": "attach",
            "droplet_id": node.id,
            "volume_name": volume.name,
            "region": volume.extra["region_slug"],
        }

        res = self.connection.request("/v2/volumes/actions", data=json.dumps(attr), method="POST")

        return res.status == httplib.ACCEPTED

    def detach_volume(self, volume):
        """
        Detaches a volume from a node.

        :param volume: Volume to be detached
        :type volume: :class:`.StorageVolume`

        :rtype: ``bool``
        """
        attr = {
            "type": "detach",
            "volume_name": volume.name,
            "region": volume.extra["region_slug"],
        }

        responses = []
        for droplet_id in volume.extra["droplet_ids"]:
            attr["droplet_id"] = droplet_id
            res = self.connection.request(
                "/v2/volumes/actions", data=json.dumps(attr), method="POST"
            )
            responses.append(res)

        return all([r.status == httplib.ACCEPTED for r in responses])

    def create_volume_snapshot(self, volume, name):
        """
        Create a new volume snapshot.

        :param volume: Volume to create a snapshot for
        :type volume: class:`StorageVolume`

        :return: The newly created volume snapshot.
        :rtype: :class:`VolumeSnapshot`
        """
        attr = {"name": name}
        res = self.connection.request(
            "/v2/volumes/%s/snapshots" % (volume.id),
            data=json.dumps(attr),
            method="POST",
        )
        data = res.object["snapshot"]
        return self._to_volume_snapshot(data=data)

    def list_volume_snapshots(self, volume):
        """
        List snapshots for a volume.

        :param volume: Volume to list snapshots for
        :type volume: class:`StorageVolume`

        :return: List of volume snapshots.
        :rtype: ``list`` of :class: `StorageVolume`
        """
        data = self._paginated_request("/v2/volumes/%s/snapshots" % (volume.id), "snapshots")
        return list(map(self._to_volume_snapshot, data))

    def delete_volume_snapshot(self, snapshot):
        """
        Delete a volume snapshot

        :param snapshot: volume snapshot to delete
        :type snapshot: class:`VolumeSnapshot`

        :rtype: ``bool``
        """
        res = self.connection.request("v2/snapshots/%s" % (snapshot.id), method="DELETE")
        return res.status == httplib.NO_CONTENT

    def ex_get_node_details(self, node_id):
        """
        Lists details of the specified server.

        :param       node_id: ID of the node which should be used
        :type        node_id: ``str``

        :rtype: :class:`Node`
        """
        data = self._paginated_request("/v2/droplets/{}".format(node_id), "droplet")
        return self._to_node(data)

    def ex_create_floating_ip(self, location):
        """
        Create new floating IP reserved to a region.

        The newly created floating IP will not be associated to a Droplet.

        See https://developers.digitalocean.com/documentation/v2/#floating-ips

        :param location: Which data center to create the floating IP in.
        :type location: :class:`.NodeLocation`

        :rtype: :class:`DigitalOcean_v2_FloatingIpAddress`
        """
        attr = {"region": location.id}
        resp = self.connection.request("/v2/floating_ips", data=json.dumps(attr), method="POST")
        return self._to_floating_ip(resp.object["floating_ip"])

    def ex_delete_floating_ip(self, ip):
        """
        Delete specified floating IP

        :param      ip: floating IP to remove
        :type       ip: :class:`DigitalOcean_v2_FloatingIpAddress`

        :rtype: ``bool``
        """
        resp = self.connection.request("/v2/floating_ips/{}".format(ip.id), method="DELETE")
        return resp.status == httplib.NO_CONTENT

    def ex_list_floating_ips(self):
        """
        List floating IPs

        :rtype: ``list`` of :class:`DigitalOcean_v2_FloatingIpAddress`
        """
        return self._to_floating_ips(self._paginated_request("/v2/floating_ips", "floating_ips"))

    def ex_get_floating_ip(self, ip):
        """
        Get specified floating IP

        :param      ip: floating IP to get
        :type       ip: ``str``

        :rtype: :class:`DigitalOcean_v2_FloatingIpAddress`
        """
        floating_ips = self.ex_list_floating_ips()
        matching_ips = [x for x in floating_ips if x.ip_address == ip]
        if not matching_ips:
            raise ValueError("Floating ip %s not found" % ip)
        return matching_ips[0]

    def ex_attach_floating_ip_to_node(self, node, ip):
        """
        Attach the floating IP to the node

        :param      node: node
        :type       node: :class:`Node`

        :param      ip: floating IP to attach
        :type       ip: ``str`` or :class:`DigitalOcean_v2_FloatingIpAddress`

        :rtype: ``bool``
        """
        data = {"type": "assign", "droplet_id": node.id}
        resp = self.connection.request(
            "/v2/floating_ips/%s/actions" % ip.ip_address,
            data=json.dumps(data),
            method="POST",
        )
        return resp.status == httplib.CREATED

    def ex_detach_floating_ip_from_node(self, node, ip):
        """
        Detach a floating IP from the given node

        Note: the 'node' object is not used in this method but it is added
        to the signature of ex_detach_floating_ip_from_node anyway so it
        conforms to the interface of the method of the same name for other
        drivers like for example OpenStack.

        :param      node: Node from which the IP should be detached
        :type       node: :class:`Node`

        :param      ip: Floating IP to detach
        :type       ip: :class:`DigitalOcean_v2_FloatingIpAddress`

        :rtype: ``bool``
        """
        data = {"type": "unassign"}
        resp = self.connection.request(
            "/v2/floating_ips/%s/actions" % ip.ip_address,
            data=json.dumps(data),
            method="POST",
        )
        return resp.status == httplib.CREATED

    def _to_node(self, data):
        extra_keys = [
            "memory",
            "vcpus",
            "disk",
            "image",
            "size",
            "size_slug",
            "locked",
            "created_at",
            "networks",
            "kernel",
            "backup_ids",
            "snapshot_ids",
            "features",
            "tags",
        ]
        if "status" in data:
            state = self.NODE_STATE_MAP.get(data["status"], NodeState.UNKNOWN)
        else:
            state = NodeState.UNKNOWN

        created = parse_date(data["created_at"])
        networks = data["networks"]
        private_ips = []
        public_ips = []
        if networks:
            for net in networks["v4"]:
                if net["type"] == "private":
                    private_ips = [net["ip_address"]]
                if net["type"] == "public":
                    public_ips = [net["ip_address"]]

        extra = {}
        for key in extra_keys:
            if key in data:
                extra[key] = data[key]
        extra["region"] = data.get("region", {}).get("name")

        # Untouched extra values, backwards compatibility
        resolve_data = data.get("image")
        if resolve_data:
            image = self._to_image(resolve_data)
        else:
            image = None

        resolve_data = extra.get("size")
        if resolve_data:
            size = self._to_size(resolve_data)
        else:
            size = None

        node = Node(
            id=data["id"],
            name=data["name"],
            state=state,
            image=image,
            size=size,
            public_ips=public_ips,
            private_ips=private_ips,
            created_at=created,
            driver=self,
            extra=extra,
        )
        return node

    def _to_image(self, data):
        extra = {
            "distribution": data["distribution"],
            "public": data["public"],
            "slug": data["slug"],
            "regions": data["regions"],
            "min_disk_size": data["min_disk_size"],
            "created_at": data["created_at"],
        }
        return NodeImage(id=data["id"], name=data["name"], driver=self, extra=extra)

    def _to_volume(self, data):
        extra = {
            "created_at": data["created_at"],
            "droplet_ids": data["droplet_ids"],
            "region": data["region"],
            "region_slug": data["region"]["slug"],
        }

        return StorageVolume(
            id=data["id"],
            name=data["name"],
            size=data["size_gigabytes"],
            driver=self,
            extra=extra,
        )

    def _to_location(self, data):
        extra = data.get("features", [])
        return NodeLocation(
            id=data["slug"], name=data["name"], country=None, extra=extra, driver=self
        )

    def _to_size(self, data):
        extra = {
            "vcpus": data["vcpus"],
            "regions": data["regions"],
            "price_monthly": data["price_monthly"],
        }
        return NodeSize(
            id=data["slug"],
            name=data["slug"],
            ram=data["memory"],
            disk=data["disk"],
            bandwidth=data["transfer"],
            price=data["price_hourly"],
            driver=self,
            extra=extra,
        )

    def _to_key_pair(self, data):
        extra = {"id": data["id"]}
        return KeyPair(
            name=data["name"],
            fingerprint=data["fingerprint"],
            public_key=data["public_key"],
            private_key=None,
            driver=self,
            extra=extra,
        )

    def _to_volume_snapshot(self, data):
        extra = {
            "created_at": data["created_at"],
            "resource_id": data["resource_id"],
            "regions": data["regions"],
            "min_disk_size": data["min_disk_size"],
        }
        return VolumeSnapshot(
            id=data["id"],
            name=data["name"],
            size=data["size_gigabytes"],
            driver=self,
            extra=extra,
        )

    def _to_floating_ips(self, obj):
        return [self._to_floating_ip(ip) for ip in obj]

    def _to_floating_ip(self, obj):
        return DigitalOcean_v2_FloatingIpAddress(
            # There is no ID, but the IP is unique so we can use that
            id=obj["ip"],
            ip_address=obj["ip"],
            node_id=obj["droplet"]["id"] if obj["droplet"] else None,
            extra={"region": obj["region"]},
            driver=self,
        )


class DigitalOcean_v2_FloatingIpAddress:
    """
    Floating IP info.
    """

    def __init__(self, id, ip_address, node_id=None, extra=None, driver=None):
        self.id = str(id)
        self.ip_address = ip_address
        self.extra = extra
        self.node_id = node_id
        self.driver = driver

    def delete(self):
        """
        Delete this floating IP

        :rtype: ``bool``
        """
        return self.driver.ex_delete_floating_ip(self)

    def __repr__(self):
        return "<DigitalOcean_v2_FloatingIpAddress: id=%s, ip_addr=%s," " driver=%s>" % (
            self.id,
            self.ip_address,
            self.driver,
        )
